"""
operation templates for package/repository/data source objects

For new format implementations, new formats, and generally any new extension,
an operation class will likely have to be defined.  While the implementations
are a bit repetitive, the design of it is intentional to ensure that any
derivative will be forced to adhere to the pkgcore internal api.

Basically it's a crappy form of zope interfaces; converting to zope.interfaces
may occur down the line if dependencies can be kept as minimal as possible.
"""

from functools import partial

from snakeoil import klass
from snakeoil.currying import pretty_docs

from ..exceptions import PkgcoreException
from ..operations import observer as _observer


class OperationError(PkgcoreException):

    def __init__(self, api, exc=None):
        self._api = api
        self._exc = exc

    def __str__(self):
        if self._exc is not None:
            return str(self._exc)
        return f"unhandled exception in {self._api}"


class base:

    __required__ = frozenset()

    UNSUPPORTED = object()
    __casting_exception__ = OperationError

    def __init__(self, disable_overrides=(), enable_overrides=()):
        self._force_disabled = frozenset(disable_overrides)
        self._force_enabled = frozenset(enable_overrides)
        self._setup_api()

    def _get_observer(self, observer=None):
        if observer is not None:
            return observer
        return _observer.null_output()

    @klass.cached_property
    def raw_operations(self):
        return frozenset(x[len("_cmd_api_"):] for x in dir(self.__class__)
                         if x.startswith("_cmd_api_"))

    @klass.cached_property
    def enabled_operations(self):
        enabled_ops = set(self._filter_disabled_commands(self.raw_operations))
        return frozenset(self._apply_overrides(enabled_ops))

    def _apply_overrides(self, ops):
        ops.update(self._force_enabled)
        ops.difference_update(self._force_disabled)
        return ops

    @staticmethod
    def _recast_exception_decorator(exc_class, name, functor, *args, **kwds):
        try:
            return functor(*args, **kwds)
        except PkgcoreException as e:
            if isinstance(e, exc_class):
                raise
            raise exc_class(name, e) from e

    def _wrap_exception(self, functor, name):
        f = partial(
            self._recast_exception_decorator, self.__casting_exception__, name, functor)
        return pretty_docs(f)

    def _setup_api(self):
        recast_exc = self.__casting_exception__
        if recast_exc is None:
            f = lambda x, y: x
        else:
            f = self._wrap_exception
        for op in self.enabled_operations:
            setattr(self, op, f(getattr(self, '_cmd_api_%s' % op), op))

    def _filter_disabled_commands(self, sequence):
        for command in sequence:
            obj = getattr(self, '_cmd_api_%s' % command, None)
            if not getattr(obj, '_is_standalone', False):
                if not hasattr(self, '_cmd_implementation_%s' % command):
                    continue
            check_f = getattr(self, '_cmd_check_support_%s' % command, None)
            if check_f is not None and not check_f():
                continue
            yield command

    def supports(self, operation_name=None, raw=False):
        if not operation_name:
            if not raw:
                return self.enabled_operations
            return self.raw_operations
        if raw:
            return operation_name in self.raw_operations
        return operation_name in self.enabled_operations

    def run_if_supported(self, operation_name, *args, **kwds):
        """invoke an operation if it's supported

        :param operation_name: operation to run if supported
        :param args: positional args passed to the operation
        :param kwds: optional args passed to the operation
        :keyword or_return: if the operation isn't supported, return this
            (if unspecified, it returns :obj:`base.UNSUPPORTED`)
        :return: Either the value of or_return, or if the operation is
            supported, the return value from that operation
        """
        kwds.setdefault('observer', getattr(self, 'observer', self._get_observer()))
        ret = kwds.pop("or_return", self.UNSUPPORTED)
        if self.supports(operation_name):
            ret = getattr(self, operation_name)(*args, **kwds)
        return ret


def is_standalone(functor):
    """decorator to mark a api operation method as containing the implementation

    This is primarily useful for commands that can contain all of the logic in
    the template class itself, rather than requiring a glue method to be
    provided
    """
    functor._is_standalone = True
    return functor



"""A pkgset hack to provide a live-rebuild equivalent.

Note: HACK. Quick proof of concept, could do with cleaning up.
"""

from ..config.hint import ConfigHint
from ..repository.util import RepositoryGroup
from .installed import VersionedInstalled


class EclassConsumerSet(VersionedInstalled):

    pkgcore_config_type = ConfigHint(
        typename='pkgset',
        types={
            'vdb': 'refs:repo',
            'repos': 'refs:repo',
            'eclasses': 'list'},
    )

    def __init__(self, vdb, repos, eclasses):
        VersionedInstalled.__init__(self, vdb)
        self.repos = RepositoryGroup(repos)
        self.eclasses = frozenset(eclasses)

    def __iter__(self):
        for atom in VersionedInstalled.__iter__(self):
            pkgs = self.repos.match(atom)
            if not pkgs:
                # pkg is installed but no longer in any repo, just ignore it.
                continue
            assert len(pkgs) == 1, 'I do not know what I am doing: %r' % (pkgs,)
            pkg = pkgs[0]
            if self.eclasses.isdisjoint(pkg.data.get('_eclasses_', ())):
                yield atom



__all__ = ("install", "uninstall", "replace", "operations")

import os
import shutil
import time
from itertools import chain

from snakeoil import compression
from snakeoil.data_source import local_source
from snakeoil.osutils import ensure_dirs, normpath, pjoin
from snakeoil.version import get_version

from .. import __title__
from ..ebuild import conditionals
from ..log import logger
from ..operations import repo as repo_ops
from .contents import ContentsFile


def update_mtime(path, timestamp=None):
    if timestamp is None:
        timestamp = time.time()
    logger.debug(f"updating vdb timestamp for {path!r}")
    try:
        os.utime(path, (timestamp, timestamp))
    except EnvironmentError as e:
        logger.error(f"failed updated vdb timestamp for {path!r}: {e}")


class install(repo_ops.install):

    def __init__(self, repo, newpkg, observer):
        base = pjoin(repo.location, newpkg.category)
        dirname = f"{newpkg.package}-{newpkg.fullver}"
        self.install_path = pjoin(base, dirname)
        self.tmp_write_path = pjoin(base, f".tmp.{dirname}")
        super().__init__(repo, newpkg, observer)

    def add_data(self, domain):
        # error checking?
        dirpath = self.tmp_write_path
        ensure_dirs(dirpath, mode=0o755, minimal=True)
        update_mtime(self.repo.location)
        rewrite = self.repo._metadata_rewrites
        for k in self.new_pkg.tracked_attributes:
            if k == "contents":
                v = ContentsFile(pjoin(dirpath, "CONTENTS"),
                                 mutable=True, create=True)
                v.update(self.new_pkg.contents)
                v.flush()
            elif k == "environment":
                data = compression.compress_data('bzip2',
                    self.new_pkg.environment.bytes_fileobj().read())
                with open(pjoin(dirpath, "environment.bz2"), "wb") as f:
                    f.write(data)
                del data
            else:
                v = getattr(self.new_pkg, k)
                if k in ('bdepend', 'depend', 'rdepend', 'idepend'):
                    s = v.slotdep_str(domain)
                elif k == 'user_patches':
                    s = '\n'.join(chain.from_iterable(files for _, files in v))
                elif not isinstance(v, str):
                    try:
                        s = ' '.join(v)
                    except TypeError:
                        s = str(v)
                else:
                    s = v
                with open(pjoin(dirpath, rewrite.get(k, k.upper())), "w", 32768) as f:
                    if s:
                        s += '\n'
                    f.write(s)

        # ebuild_data is the actual ebuild- no point in holding onto
        # it for built ebuilds, but if it's there, we store it.
        o = getattr(self.new_pkg, "ebuild", None)
        if o is None:
            logger.warning(
                "doing install/replace op, "
                "but source package doesn't provide the actual ebuild data.  "
                "Creating an empty file")
            o = ''
        else:
            o = o.bytes_fileobj().read()
        # XXX lil hackish accessing PF
        with open(pjoin(dirpath, self.new_pkg.PF + ".ebuild"), "wb") as f:
            f.write(o)

        # install NEEDED and NEEDED.ELF.2 files from tmpdir if they exist
        pkg_tmpdir = normpath(pjoin(domain.pm_tmpdir, self.new_pkg.category,
                                    self.new_pkg.PF, 'temp'))
        for f in ['NEEDED', 'NEEDED.ELF.2']:
            fp = pjoin(pkg_tmpdir, f)
            if os.path.exists(fp):
                local_source(fp).transfer_to_path(pjoin(dirpath, f))

        # XXX finally, hack to keep portage from doing stupid shit.
        # relies on counter to discern what to punt during
        # merging/removal, we don't need that crutch however. problem?
        # No counter file, portage wipes all of our merges (friendly
        # bugger).
        # need to get zmedico to localize the counter
        # creation/counting to per CP for this trick to behave
        # perfectly.
        with open(pjoin(dirpath, "COUNTER"), "w") as f:
            f.write(str(int(time.time())))

        # finally, we mark who made this.
        with open(pjoin(dirpath, "PKGMANAGER"), "w") as f:
            f.write(get_version(__title__, __file__))
        return True

    def finalize_data(self):
        os.rename(self.tmp_write_path, self.install_path)
        update_mtime(self.repo.location)
        return True


class uninstall(repo_ops.uninstall):

    def __init__(self, repo, pkg, observer):
        self.remove_path = pjoin(
            repo.location, pkg.category, pkg.package+"-"+pkg.fullver)
        super().__init__(repo, pkg, observer)

    def remove_data(self):
        return True

    def finalize_data(self):
        update_mtime(self.repo.location)
        shutil.rmtree(self.remove_path)
        update_mtime(self.repo.location)
        return True


# should convert these to mixins.
class replace(repo_ops.replace, install, uninstall):

    def __init__(self, repo, pkg, newpkg, observer):
        uninstall.__init__(self, repo, pkg, observer)
        install.__init__(self, repo, newpkg, observer)

    remove_data = uninstall.remove_data

    def add_data(self, domain):
        return install.add_data(self, domain)

    def finalize_data(self):
        # XXX: should really restructure this into
        # a rename of the unmerge dir, rename merge into it's place (for
        # literal same fullver replacements), then wipe the unmerge
        # that minimizes the window for races, and gets the data in place
        # should unmerge somehow die.
        uninstall.finalize_data(self)
        install.finalize_data(self)
        return True


class operations(repo_ops.operations):

    def _cmd_implementation_install(self, pkg, observer):
        return install(self.repo, pkg, observer)

    def _cmd_implementation_uninstall(self, pkg, observer):
        return uninstall(self.repo, pkg, observer)

    def _cmd_implementation_replace(self, oldpkg, newpkg, observer):
        return replace(self.repo, oldpkg, newpkg, observer)

    def _cmd_api_regen_cache(self, *args, **kwargs):
        # disable threaded cache updates
        super()._cmd_api_regen_cache(*args, threads=1, **kwargs)



"""
virtual repository, pkgs generated via callable
"""

__all__ = ("tree", "RestrictionRepo")

from snakeoil.compatibility import cmp
from snakeoil.sequences import stable_unique

from ..ebuild import atom
from ..ebuild.conditionals import DepSet
from ..package import base as pkg_base
from ..package import virtual
from ..restrictions.boolean import OrRestriction
from . import prototype


class tree(prototype.tree):

    factory_kls = staticmethod(virtual.factory)

    def __init__(self, livefs=False, frozen=False):
        """
        :param grab_virtuals_func: callable to get a package -> versions mapping
        :param livefs: is this a livefs repository?
        """
        super().__init__(frozen)
        self.livefs = livefs
        vf = self.factory_kls(self)
        self.package_class = vf.new_package

    def _expand_vers(self, cp, ver):
        raise NotImplementedError(self, "_expand_vers")

    def _internal_gen_candidates(self, candidates, sorter):
        pkls = self.package_class
        for cp in candidates:
            for pkg in sorter(
                    pkls(provider, cp[0], cp[1], ver)
                    for ver in self.versions.get(cp, ())
                    for provider in self._expand_vers(cp, ver)):
                yield pkg

    def _get_categories(self, *optional_category):
        # return if optional_category is passed... cause it's not yet supported
        if optional_category:
            return ()
        return ("virtual",)

    def _load_data(self):
        raise NotImplementedError(self, "_load_data")

    def _get_packages(self, category):
        if category != 'virtual':
            raise KeyError(f'no {category} category for this repository')
        self._load_data()
        return self.packages[category]


class InjectedPkg(pkg_base.wrapper):

    __slots__ = (
        "bdepend", "depend", "rdepend", "pdepend", "idepend",
        "repo", "repo_id", "built", "versioned_atom", "unversioned_atom", "data",
    )
    default_bdepend = default_depend = default_rdepend = default_pdepend = default_idepend = DepSet()
    package_is_real = False
    is_supported = True

    def __init__(self, raw_pkg, repo, data=None):
        pkg_base.wrapper.__init__(self, raw_pkg)
        object.__setattr__(self, "repo", repo)
        object.__setattr__(self, "repo_id", repo.repo_id)
        object.__setattr__(self, "built", repo.livefs)
        object.__setattr__(self, "versioned_atom", self._raw_pkg)
        object.__setattr__(self, "unversioned_atom", self._raw_pkg.key)
        object.__setattr__(self, "bdepend", self.default_bdepend)
        object.__setattr__(self, "depend", self.default_depend)
        object.__setattr__(self, "rdepend", self.default_rdepend)
        object.__setattr__(self, "pdepend", self.default_pdepend)
        object.__setattr__(self, "idepend", self.default_idepend)
        object.__setattr__(self, "data", data)

    @property
    def use(self):
        if self._raw_pkg.use is None:
            return ()
        return self._raw_pkg.use

    def __cmp__(self, other):
        if isinstance(other, InjectedPkg):
            other = other._raw_pkg
        elif isinstance(other, pkg_base.base):
            other = other.versioned_atom
        if self._raw_pkg.intersects(other):
            return 0
        return cmp(self._raw_pkg, other)

    def __eq__(self, other):
        if isinstance(other, InjectedPkg):
            other = other._raw_pkg
        elif isinstance(other, pkg_base.base):
            other = other.versioned_atom
        return self._raw_pkg.intersects(other)

    def __ne__(self, other):
        if isinstance(other, InjectedPkg):
            other = other._raw_pkg
        elif isinstance(other, pkg_base.base):
            other = other.versioned_atom
        return not self._raw_pkg.intersects(other)

    def __str__(self):
        return f'injected restriction pkg: {self._raw_pkg}'

    def __repr__(self):
        return "<%s cpv=%r @%#8x>" % (self.__class__, self.cpvstr, id(self))

    def __hash__(self):
        return hash(self._raw_pkg)


class RestrictionRepo(tree):
    """Fake repo populated by packages matching a given restriction."""

    def __init__(self, repo_id, restrictions=(), frozen=False, livefs=False):
        self.repo_id = repo_id
        self._injected_pkgs = {}
        self._restrictions = {}
        for r in restrictions:
            self[r] = None
        super().__init__(livefs=livefs, frozen=frozen)

    def __setitem__(self, key, val):
        if isinstance(key, atom.atom):
            self._injected_pkgs[InjectedPkg(key, self, data=val)] = val
        else:
            self._restrictions[key] = val

    def __getitem__(self, key):
        val = self._injected_pkgs.__getitem__(key)
        if val is not None:
            return val
        return self._restrictions.__getitem__(key)

    @property
    def restriction(self):
        return OrRestriction(*self._restrictions.keys())

    def _get_categories(self, *args):
        return tuple(x.category for x in self._injected_pkgs)

    def _get_packages(self, category):
        return tuple(x.package for x in self._injected_pkgs)

    def _get_versions(self, package):
        return tuple(x.version for x in self._injected_pkgs)

    def __iter__(self):
        return iter(self._injected_pkgs)

    def itermatch(self, restrict, sorter=iter, pkg_cls=InjectedPkg):
        if isinstance(restrict, atom.atom):
            func = restrict.intersects
        else:
            func = restrict.match

        # yield any matching pkgs already injected into the repo
        for pkg in self._injected_pkgs:
            if func(pkg._raw_pkg):
                yield pkg

        # inject/yield any matching atoms into the repo that aren't blockers
        if self._restrictions and isinstance(restrict, atom.atom):
            if self.restriction.match(restrict) and not restrict.blocks:
                p = pkg_cls(restrict, self)
                self._injected_pkgs[p] = None
                yield p

    def match(self, restrict, **kwargs):
        return list(self.itermatch(restrict, **kwargs))



import io
from functools import partial

from pkgcore.ebuild import filter_env
from snakeoil.test import TestCase


class TestFilterEnv(TestCase):

    filter_env = staticmethod(partial(filter_env.main_run))

    def get_output(self, raw_data, funcs=None, vars=None, preserve_funcs=False,
                   preserve_vars=False, debug=False, global_envvar_callback=None):
        out = io.BytesIO()
        if funcs:
            funcs = funcs.split(',')
        if vars:
            vars = vars.split(',')
        self.filter_env(out, raw_data, vars, funcs, preserve_vars, preserve_funcs,
            global_envvar_callback=global_envvar_callback)
        return out.getvalue().decode('utf-8')

    def test_function_foo(self):
        ret = ''.join(self.get_output("function foo() {:;}", funcs="foo"))
        self.assertEqual(ret, '')
        ret = ''.join(self.get_output("functionfoo() {:;}", funcs="foo"))
        self.assertEqual(ret, 'functionfoo() {:;}')

    def test_simple(self):
        data = \
"""
foo() {
    :
}

bar() {
    :
}
"""
        ret = ''.join(self.get_output(data))
        self.assertIn('foo', ret)
        self.assertIn('bar', ret)
        ret = ''.join(self.get_output(data, funcs='foo'))
        self.assertNotIn('foo', ret)
        self.assertIn('bar', ret)
        ret = ''.join(self.get_output(data, funcs='bar'))
        self.assertIn('foo', ret)
        self.assertNotIn('bar', ret)
        ret = ''.join(self.get_output(data, funcs='bar,foo'))
        self.assertNotIn('foo', ret)
        self.assertNotIn('bar', ret)

    def test1(self):
        data = \
"""
MODULE_NAMES=${MODULE_NAMES//${i}(*};
tc-arch ()
{
    tc-ninja_magic_to_arch portage $@
}
"""
        ret = ''.join(self.get_output(data, vars='MODULE_NAMES'))
        self.assertNotIn('MODULE_NAMES', ret)
        self.assertIn('tc-arch', ret)

    def test_comments(self):
        data = "dar=${yar##.%}\nfoo() {\n:\n}\n"
        ret = ''.join(self.get_output(data, vars='dar'))
        self.assertNotIn('dar', ret)
        self.assertIn('foo', ret)

        data = \
"""
src_unpack() {
    use idn && {
       # BIND 9.4.0 doesn't have this patch
       :
    }
}

src_compile() {
    :
}
"""
        self.assertIn('src_unpack', ''.join(
            self.get_output(data, funcs='src_compile')))
        ret = ''.join(self.get_output(data, funcs='src_unpack'))
        self.assertIn('src_compile', ret)
        self.assertNotIn('src_unpack', ret)
        data = \
"""src_install() {
    local -f ${f##*=}
}

pkg_postinst() {
    :
}
"""
        self.assertNotIn('pkg_postinst',
            ''.join(self.get_output(data, funcs='pkg_postinst')))
        data = \
"""src_unpack() {
    fnames=$(scanelf -pyqs__uClibc_start_main -F%F#s)
}
src_compile() {
    :
}
"""
        self.assertIn('src_compile',
            ''.join(self.get_output(data, funcs='src_unpack')))

        data = \
"""findtclver() {
    [ "$(#i)" = "3" ]
}

pkg_setup() {
    :
}
"""
        self.assertIn('pkg_setup',
            ''.join(self.get_output(data, funcs='findtclver')))

    def test_here(self):
        data = \
"""
src_install() {
    cat >${D}/etc/modules.d/davfs2 <<EOF
alias char-major-67 coda
alias /dev/davfs*   coda
EOF
}

pkg_setup() {
    :
}
"""
        self.assertNotIn('pkg_setup', ''.join(self.get_output(data,
            funcs='pkg_setup')))

        data = \
"""
pkg_setup() {
        while read line; do elog "${line}"; done <<EOF
The default behaviour of tcsh has significantly changed starting from
version 6.14-r1.  In contrast to previous ebuilds, the amount of
customisation to the default shell's behaviour has been reduced to a
bare minimum (a customised prompt).
If you rely on the customisations provided by previous ebuilds, you will
have to copy over the relevant (now commented out) parts to your own
~/.tcshrc.  Please check all tcsh-* files in
/usr/share/doc/${P}/examples/ and include their behaviour in your own
configuration files.
The tcsh-complete file is not any longer sourced by the default system
scripts.
EOF
}

pkg_foo() {
    :
}
"""
        self.assertNotIn('pkg_foo', ''.join(self.get_output(data,
            funcs='pkg_foo')))

    def test_vars(self):
        data = \
"""
f() {
    x=$y
}

z() {
    :
}
"""
        self.assertIn('z', ''.join(self.get_output(data,
            funcs='f')))

        data = \
"""
f() {
    x="${y}"
}

z() {
    :
}
"""
        self.assertIn('z', ''.join(self.get_output(data,
            funcs='f')))

        data = \
"""src_compile() {
    $(ABI=foo get_libdir)
}

pkg_setup() {
    :
}
"""
        self.assertIn('pkg_setup', ''.join(self.get_output(data,
            funcs='src_compile')))

    def test_quoting(self):
        data = \
"""
pkg_postinst() {
    einfo " /bin/ls ${ROOT}etc/init.d/net.* | grep -v '/net.lo$' | xargs -n1 ln -sfvn net.lo"
}

pkg_setup() {
    :
}
"""
        self.assertIn('pkg_setup', ''.join(self.get_output(data,
            funcs='pkg_postinst')))

        data = \
"""src_unpack() {
    testExp=$'\177\105\114\106\001\001\001'
}

src_install() {
    :
}
"""
        self.assertIn('src_install', ''.join(self.get_output(data,
            funcs='src_unpack')))

    def test_arg_awareness(self):
        data = "f() {\n x \\{}\n}\n"
        self.assertNotIn('}', ''.join(self.get_output(data, 'f')))

    def test_print_vars(self):
        def assertVars(data, var_list, assert_func=self.assertEqual):
            l = []
            self.get_output(data, global_envvar_callback=l.append)
            assert_func(sorted(var_list), sorted(l))
        assertVars("f(){\nX=dar\n}", [])
        assertVars("f(){\nX=dar\n}\nY=a", ['Y'])
        assertVars("f(){\nX=dar\n}\nmy command\nY=a\nf=$(dar)", ['Y', 'f'])
        assertVars("f(){\nX=dar\n}\nmy command\nY=a\nf=$(dar) foon\n", ['Y'],
            self.assertNotEqual)
        assertVars("f(){\nX=dar foon\n}\nY=dar\nf2(){Z=dar;}\n", ['Y'])



import logging
import os
import re

import pytest
from pkgcore.ebuild import atom, repo_objs
from pkgcore.repository import errors as repo_errors
from snakeoil.data_source import data_source
from snakeoil.fileutils import touch
from snakeoil.mappings import ImmutableDict


class TestMetadataXml:

    @staticmethod
    def get_metadata_xml(maintainers=(), comments=(), local_use={},
                         longdescription=None, maint_type=None,
                         proxied=None, stabilize_allarches=False):
        cs = '\n'.join(comments)
        ms = us = ls = ""
        if maintainers:
            ms = []
            for x in maintainers:
                ms.append(f"<email>{x[0]}</email>")
                if len(x) > 1:
                    ms[-1] += f"\n<name>{x[1]}</name>"
                if len(x) > 2:
                    ms[-1] += f"\n<description>{x[2]}</description>"
                if len(x) > 3:
                    raise ValueError('maintainer data has too many fields')
            maint_type = f'type="{maint_type}"' if maint_type is not None else ''
            proxied = f'proxied="{proxied}"' if proxied is not None else ''
            ms = '\n'.join(f'<maintainer {maint_type} {proxied}>{x}</maintainer>'
                           for x in ms)
        if local_use:
            us = ['<use>']
            for flag, desc in local_use.items():
                us.append(f'<flag name="{flag}">{desc}</flag>')
            us.append('</use>')
            us = '\n'.join(us)
        if longdescription:
            ls = f"<longdescription>{longdescription}</longdescription>\n"
        sa = "<stabilize-allarches/>" if stabilize_allarches else ""
        s = \
f"""<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE pkgmetadata SYSTEM "http://www.gentoo.org/dtd/metadata.dtd">
<pkgmetadata>
{cs}{ms}{us}{ls}{sa}</pkgmetadata>"""
        return repo_objs.MetadataXml(data_source(s.encode('utf-8')))

    def test_empty_maintainers(self):
        assert () == self.get_metadata_xml().maintainers

    def test_maintainer_needed(self):
        mx = self.get_metadata_xml(comments=('<!-- maintainer-needed -->',))
        assert mx.maintainers == ()

    def test_multiple_maintainers(self):
        names = ("foo@gmail.com", "monkeybone@gmail.com")
        mx = self.get_metadata_xml(maintainers=tuple((x,) for x in names))
        assert sorted(names) == sorted(map(str, mx.maintainers))
        assert "foo@gmail.com" in mx.maintainers
        assert "monkeybone@gmail.com" in mx.maintainers

    def test_maintainer_name_with_email(self):
        mx = self.get_metadata_xml(
            maintainers=(("funkymonkey@gmail.com", "funky monkey \N{SNOWMAN}"),))
        assert ("funky monkey \N{SNOWMAN} <funkymonkey@gmail.com>",) == \
            tuple(map(str, mx.maintainers))
        assert "funkymonkey@gmail.com" in mx.maintainers
        assert "funky monkey \N{SNOWMAN}" in mx.maintainers
        assert "funkymonkey@gmail.com" == mx.maintainers[0].email
        assert "funky monkey \N{SNOWMAN}" == mx.maintainers[0].name
        assert mx.maintainers[0].description is None
        assert mx.maintainers[0].maint_type is None
        assert mx.maintainers[0].proxied is None

    def test_maintainer_with_desc(self):
        mx = self.get_metadata_xml(
            maintainers=(("foo@bar.com", "foobar", "Foobar"),))
        assert ("foobar <foo@bar.com> (Foobar)",) == tuple(map(str, mx.maintainers))
        assert "foo@bar.com" in mx.maintainers
        assert "foobar" in mx.maintainers
        assert "foo@bar.com" == mx.maintainers[0].email
        assert "foobar" == mx.maintainers[0].name
        assert "Foobar" == mx.maintainers[0].description
        assert mx.maintainers[0].maint_type is None
        assert mx.maintainers[0].proxied is None

    def test_maintainer_with_type(self):
        mx = self.get_metadata_xml(
            maintainers=(("foo@bar.com", "foobar"),),
            maint_type='person')
        assert ("foobar <foo@bar.com>",) == tuple(map(str, mx.maintainers))
        assert "foo@bar.com" in mx.maintainers
        assert "foobar" in mx.maintainers
        assert "foo@bar.com" == mx.maintainers[0].email
        assert "foobar" == mx.maintainers[0].name
        assert mx.maintainers[0].description is None
        assert "person" == mx.maintainers[0].maint_type
        assert mx.maintainers[0].proxied is None

    def test_maintainer_with_proxied(self):
        mx = self.get_metadata_xml(
            maintainers=(("foo@bar.com", "foobar"),),
            maint_type='person',
            proxied='yes')
        assert ("foobar <foo@bar.com>",) == tuple(map(str, mx.maintainers))
        assert "foo@bar.com" in mx.maintainers
        assert "foobar" in mx.maintainers
        assert "foo@bar.com" == mx.maintainers[0].email
        assert "foobar" == mx.maintainers[0].name
        assert mx.maintainers[0].description is None
        assert "person" == mx.maintainers[0].maint_type
        assert "yes" == mx.maintainers[0].proxied

    def test_local_use(self):
        # empty...
        assert dict() == self.get_metadata_xml().local_use

        local_use = {
            "foo": "description for foo",
            "bar": "description for bar (<pkg>app-foo/bar</pkg> required)",
        }
        metadata_xml = self.get_metadata_xml(local_use=local_use)
        pkg_tag_re = re.compile(r'</?pkg>')
        local_use = dict(
            (k, pkg_tag_re.sub('', v))
            for k, v in local_use.items())
        assert local_use == metadata_xml.local_use

    def test_longdesc(self):
        # empty...
        assert None == self.get_metadata_xml().longdescription
        s = \
"""
I saw the best minds of my generation destroyed by madness, starving
hysterical naked, dragging themselves throughout the negro streets at dawn
looking for an angry fix, angle-headed hipsters burning for the ancient
heavenly connection to the starry dynamo in the machinery of night, who
poverty and tatters and hollowed-eyed and high sat up smoking in the
supernatural darkness of cold-water flats floating across the tops of cities
contemplating jazz, who bared their brains to Heaven under the El and saw
Mohammedan angels staggering on tenement roofs illuminated, who passed
through universities with radiant cool eyes hallucinating Arkansas and
Blake-light tragedy among the scholars of war.
"""

        assert " ".join(s.split()) == self.get_metadata_xml(longdescription=s).longdescription

    def test_stabilize_allarches(self):
        # missing
        assert False == self.get_metadata_xml().stabilize_allarches
        # present
        assert True == self.get_metadata_xml(stabilize_allarches=True).stabilize_allarches


class TestProjectsXml:

    @staticmethod
    def get_projects_xml(s=None):
        default_s = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE projects SYSTEM "http://www.gentoo.org/dtd/projects.dtd">
<projects>
  <project>
    <email>nomembers@example.com</email>
    <name>Project with no members</name>
    <url>https://projects.example.com/nomembers</url>
    <description>Here is the description</description>
  </project>
  <project>
    <email>nolead@example.com</email>
    <name>Project with no lead</name>
    <url>https://projects.example.com/nolead</url>
    <description>Here is the description</description>
    <member>
      <email>fulldev1@example.com</email>
      <name>Full Dev</name>
      <role>Somebody</role>
    </member>
    <member>
      <email>dev1@example.com</email>
    </member>
  </project>
  <project>
    <email>lead@example.com</email>
    <name>Project with a lead</name>
    <url>https://projects.example.com/lead</url>
    <description>Here is the description</description>
    <member>
      <email>fulldev2@example.com</email>
      <name>Full Dev</name>
      <role>Somebody</role>
    </member>
    <member is-lead="1">
      <email>dev2@example.com</email>
    </member>
  </project>
  <project>
    <email>subprojects@example.com</email>
    <name>Project with non-recursive subprojects</name>
    <url>https://projects.example.com/subprojects</url>
    <description>Here is the description</description>
    <member>
      <email>fulldev3@example.com</email>
      <name>Full Dev</name>
      <role>Somebody</role>
    </member>
    <member is-lead="1">
      <email>dev3@example.com</email>
    </member>
    <subproject ref="nolead@example.com"/>
    <subproject ref="lead@example.com"/>
  </project>
  <project>
    <email>recursive-subprojects@example.com</email>
    <name>Project with recursive subprojects</name>
    <url>https://projects.example.com/recursive-subprojects</url>
    <description>Here is the description</description>
    <member>
      <email>fulldev4@example.com</email>
      <name>Full Dev</name>
      <role>Somebody</role>
    </member>
    <member>
      <!-- this one is common -->
      <email>dev1@example.com</email>
    </member>
    <member is-lead="1">
      <email>dev4@example.com</email>
    </member>
    <subproject ref="nolead@example.com" inherit-members="1"/>
    <subproject ref="lead@example.com"/>
  </project>
  <project>
    <email>super-recursive-subprojects@example.com</email>
    <name>Project with double recursion</name>
    <url>https://projects.example.com/super-recursive-subprojects</url>
    <description>Here is the description</description>
    <member>
      <email>fulldev5@example.com</email>
      <name>Full Dev</name>
      <role>Somebody</role>
    </member>
    <member is-lead="1">
      <!-- this one is common -->
      <email>dev2@example.com</email>
    </member>
    <member>
      <email>dev5@example.com</email>
    </member>
    <subproject ref="recursive-subprojects@example.com" inherit-members="1"/>
  </project>
</projects>"""
        if s is None:
            s = default_s
        return repo_objs.ProjectsXml(data_source(s.encode('utf-8')))

    def test_empty(self):
        assert self.get_projects_xml('').projects == {}

    def test_invalid_xml(self):
        assert self.get_projects_xml('<foo><bar></foo>').projects == {}

    def test_project_member_without_email(self):
        p = self.get_projects_xml("""<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE projects SYSTEM "http://www.gentoo.org/dtd/projects.dtd">
<projects>
  <project>
    <email>nolead@example.com</email>
    <name>Project with no lead</name>
    <url>https://projects.example.com/nolead</url>
    <description>Here is the description</description>
    <member>
      <name>Full Dev</name>
      <role>Somebody</role>
    </member>
  </project>
</projects>
""").projects['nolead@example.com']
        assert p.members == ()

    def test_subproject_without_ref(self):
        p = self.get_projects_xml("""<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE projects SYSTEM "http://www.gentoo.org/dtd/projects.dtd">
<projects>
  <project>
    <email>nolead@example.com</email>
    <name>Project with no lead</name>
    <url>https://projects.example.com/nolead</url>
    <description>Here is the description</description>
    <member>
      <email>fulldev1@example.com</email>
      <name>Full Dev</name>
      <role>Somebody</role>
    </member>
    <subproject inherit-members="1"/>
  </project>
</projects>
""").projects['nolead@example.com']
        assert p.subprojects == ()

    def test_subproject_with_invalid_ref(self):
        p = self.get_projects_xml("""<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE projects SYSTEM "http://www.gentoo.org/dtd/projects.dtd">
<projects>
  <project>
    <email>nolead@example.com</email>
    <name>Project with no lead</name>
    <url>https://projects.example.com/nolead</url>
    <description>Here is the description</description>
    <member>
      <email>fulldev1@example.com</email>
      <name>Full Dev</name>
      <role>Somebody</role>
    </member>
    <subproject ref="nonexist@example.com" inherit-members="1"/>
  </project>
</projects>
""").projects['nolead@example.com']
        assert p.subprojects[0].project is None
        assert p.recursive_members == p.members

    def test_deep_subproject_with_invalid_ref(self):
        p = self.get_projects_xml("""<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE projects SYSTEM "http://www.gentoo.org/dtd/projects.dtd">
<projects>
  <project>
    <email>nolead@example.com</email>
    <name>Project with no lead</name>
    <url>https://projects.example.com/nolead</url>
    <description>Here is the description</description>
    <member>
      <email>fulldev1@example.com</email>
      <name>Full Dev</name>
      <role>Somebody</role>
    </member>
    <subproject ref="nonexist@example.com" inherit-members="1"/>
  </project>
  <project>
    <email>subprojects@example.com</email>
    <name>Project with real subprojects</name>
    <url>https://projects.example.com/subprojects</url>
    <description>Here is the description</description>
    <subproject ref="nolead@example.com" inherit-members="1"/>
  </project>
</projects>
""").projects
        assert (p['subprojects@example.com'].recursive_members ==
                p['nolead@example.com'].members)

    def test_basic_metadata(self):
        projects = self.get_projects_xml().projects
        for email, project in projects.items():
            assert project.email == email
            assert project.name.startswith('Project with')
            assert project.url == (
                'https://projects.example.com/' + email.split('@')[0])
            assert project.description == 'Here is the description'

    def test_no_members(self):
        p = self.get_projects_xml().projects['nomembers@example.com']
        assert p.members == ()
        assert p.leads == ()
        assert p.subprojects == ()
        assert p.recursive_members == ()

    def test_no_lead(self):
        p = self.get_projects_xml().projects['nolead@example.com']
        assert p.members[0].email == 'fulldev1@example.com'
        assert p.members[0].name == 'Full Dev'
        assert p.members[0].role == 'Somebody'
        assert not p.members[0].is_lead
        assert p.members[1].email == 'dev1@example.com'
        assert p.members[1].name is None
        assert p.members[1].role is None
        assert not p.members[1].is_lead
        assert p.leads == ()
        assert p.subprojects == ()
        assert p.recursive_members == p.members

    def test_have_lead(self):
        p = self.get_projects_xml().projects['lead@example.com']
        assert p.members[0].email == 'fulldev2@example.com'
        assert p.members[0].name == 'Full Dev'
        assert p.members[0].role == 'Somebody'
        assert not p.members[0].is_lead
        assert p.members[1].email == 'dev2@example.com'
        assert p.members[1].name is None
        assert p.members[1].role is None
        assert p.members[1].is_lead
        assert p.leads == (p.members[1],)
        assert p.subprojects == ()
        assert p.recursive_members == p.members

    def test_subprojects(self):
        p = self.get_projects_xml().projects['subprojects@example.com']
        assert p.subprojects[0].email == 'nolead@example.com'
        assert p.subprojects[0].name == 'Project with no lead'
        assert not p.subprojects[0].inherit_members
        assert p.subprojects[1].email == 'lead@example.com'
        assert p.subprojects[1].name == 'Project with a lead'
        assert not p.subprojects[1].inherit_members
        assert p.recursive_members == p.members

    @staticmethod
    def unlead(members):
        for m in members:
            yield repo_objs.ProjectMember(
                email=m.email, name=m.name, role=m.role, is_lead=False)

    def test_recursive_subprojects(self):
        projects = self.get_projects_xml().projects
        p = projects['recursive-subprojects@example.com']
        sp = projects['nolead@example.com']
        assert p.subprojects[0].email == 'nolead@example.com'
        assert p.subprojects[0].name == 'Project with no lead'
        assert p.subprojects[0].inherit_members
        assert p.subprojects[1].email == 'lead@example.com'
        assert p.subprojects[1].name == 'Project with a lead'
        assert not p.subprojects[1].inherit_members
        # one extra member should be inherited from sp
        assert p.recursive_members == (
            p.members + tuple(self.unlead(sp.members[:1])))
        # lead of sp should not be considered lead of p
        assert not p.recursive_members[3].is_lead

    def test_super_recursive_subprojects(self):
        projects = self.get_projects_xml().projects
        p = projects['super-recursive-subprojects@example.com']
        sp = projects['recursive-subprojects@example.com']
        ssp = projects['nolead@example.com']
        assert p.subprojects[0].email == 'recursive-subprojects@example.com'
        assert p.subprojects[0].name == 'Project with recursive subprojects'
        assert p.subprojects[0].inherit_members
        # one extra member should be inherited from ssp
        assert p.recursive_members == (
            p.members + tuple(self.unlead(sp.members + ssp.members[:1])))
        # lead of sp should not be considered lead of p
        assert not p.recursive_members[5].is_lead


class TestRepoConfig:

    @pytest.fixture(autouse=True)
    def _setup(self, tmpdir):
        self.repo_path = str(tmpdir)
        self.profiles_base = os.path.join(self.repo_path, 'profiles')
        self.metadata_path = os.path.join(
            self.repo_path, repo_objs.RepoConfig.layout_offset)
        self.eapi_path = os.path.join(self.profiles_base, 'eapi')

    def test_nonexistent_repo(self):
        # Newly configured, nonexistent repos shouldn't cause issues.
        repo_config = repo_objs.RepoConfig('nonexistent')
        assert repo_config.location == 'nonexistent'

    def test_default_eapi(self):
        os.mkdir(self.profiles_base)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert str(repo_config.eapi) == '0'

    def test_empty_file_eapi(self):
        os.mkdir(self.profiles_base)
        touch(self.eapi_path)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert str(repo_config.eapi) == '0'

    def test_empty_content_eapi(self):
        os.mkdir(self.profiles_base)
        with open(self.eapi_path, 'w+') as f:
            f.write('     \n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert str(repo_config.eapi) == '0'

    def test_unknown_eapi(self):
        os.mkdir(self.profiles_base)
        with open(self.eapi_path, 'w+') as f:
            f.write('unknown_eapi')
        with pytest.raises(repo_errors.UnsupportedRepo) as excinfo:
            repo_objs.RepoConfig(self.repo_path)
        assert isinstance(excinfo.value.repo, repo_objs.RepoConfig)

    def test_known_eapi(self):
        os.mkdir(self.profiles_base)
        with open(self.eapi_path, 'w+') as f:
            f.write('6')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert str(repo_config.eapi) == '6'

    def test_bad_data_known_eapi(self, caplog):
        os.mkdir(self.profiles_base)
        with open(self.eapi_path, 'w+') as f:
            f.write('4\nfoo\nbar')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert str(repo_config.eapi) == '4'
        assert 'multiple lines detected' in caplog.text

    def test_bad_data_unknown_eapi(self, caplog):
        os.mkdir(self.profiles_base)
        with open(self.eapi_path, 'w+') as f:
            f.write('eapi\nfoo\nbar')
        with pytest.raises(repo_errors.UnsupportedRepo) as excinfo:
            repo_objs.RepoConfig(self.repo_path)
        assert isinstance(excinfo.value.repo, repo_objs.RepoConfig)
        assert 'multiple lines detected' in caplog.text

    def test_is_empty(self, caplog):
        caplog.set_level(logging.DEBUG)

        # nonexistent repo
        repo_config = repo_objs.RepoConfig('nonexistent')
        assert repo_config.is_empty
        assert caplog.text == ''
        caplog.clear()
        del repo_config

        # empty repo
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.is_empty
        assert 'repo is empty:' in caplog.text
        caplog.clear()
        del repo_config

        # profiles dir exists
        os.mkdir(self.profiles_base)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert not repo_config.is_empty
        del repo_config

    def test_repo_name(self, caplog):
        # nonexistent file
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.repo_name is None
        del repo_config

        # empty file
        os.mkdir(os.path.dirname(self.metadata_path))
        touch(self.metadata_path)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.repo_name is None
        del repo_config

        # bad data formatting
        with open(self.metadata_path, 'w') as f:
            f.write('repo-name repo')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.repo_name is None
        assert 'bash parse error' in caplog.text
        caplog.clear()
        del repo_config

        # bad data formatting + name
        with open(self.metadata_path, 'w') as f:
            f.write('foo bar\nrepo-name = repo0')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.repo_name == 'repo0'
        assert 'bash parse error' in caplog.text
        caplog.clear()
        del repo_config

        # unset
        with open(self.metadata_path, 'w') as f:
            f.write('repo-name =')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.repo_name == ''
        del repo_config

        # whitespace
        with open(self.metadata_path, 'w') as f:
            f.write('repo-name =  \n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.repo_name == ''
        del repo_config

        # whitespace + name
        with open(self.metadata_path, 'w') as f:
            f.write('repo-name = repo \n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.repo_name == 'repo'
        del repo_config

        # regular name
        with open(self.metadata_path, 'w') as f:
            f.write('repo-name = repo1')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.repo_name == 'repo1'
        del repo_config

    def test_manifests(self):
        # nonexistent file
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.manifests == {
            'disabled': False,
            'strict': True,
            'thin': False,
            'signed': True,
            'hashes': repo_objs.RepoConfig.default_hashes,
            'required_hashes': repo_objs.RepoConfig.default_required_hashes,
        }
        del repo_config

        # regular data
        os.mkdir(os.path.dirname(self.metadata_path))
        with open(self.metadata_path, 'w') as f:
            f.write('manifest-hashes = foo\n')
            f.write('manifest-required-hashes = bar\n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.manifests.hashes == ('size', 'foo')
        assert repo_config.manifests.required_hashes == ('size', 'bar')
        del repo_config

    def test_masters(self, caplog):
        # empty repo
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.masters == ()
        assert caplog.text == ''
        caplog.clear()
        del repo_config

        # nonempty repo
        os.mkdir(self.profiles_base)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.masters == ()
        assert "doesn't specify masters in metadata" in caplog.text
        caplog.clear()
        del repo_config

        # explicit empty masters for standalone repo
        os.mkdir(os.path.dirname(self.metadata_path))
        with open(self.metadata_path, 'w') as f:
            f.write('masters =\n')
        repo_config = repo_objs.RepoConfig(self.repo_path, config_name='foo')
        assert repo_config.masters == ()
        assert caplog.text == ''
        caplog.clear()
        del repo_config

        # overlay repo with masters
        with open(self.metadata_path, 'w') as f:
            f.write('masters = foo bar\n')
        repo_config = repo_objs.RepoConfig(self.repo_path, config_name='a')
        assert repo_config.masters == ('foo', 'bar')
        del repo_config

        # overlay repo with duplicate masters
        with open(self.metadata_path, 'w') as f:
            f.write('masters = foo bar foo baz\n')
        repo_config = repo_objs.RepoConfig(self.repo_path, config_name='b')
        assert repo_config.masters == ('foo', 'bar', 'baz')
        del repo_config

    def test_cache_format(self, caplog):
        # empty repo
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.cache_format == 'md5-dict'
        del repo_config

        # explicit empty setting
        os.mkdir(os.path.dirname(self.metadata_path))
        with open(self.metadata_path, 'w') as f:
            f.write('cache-formats =\n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.cache_format is None
        del repo_config

        # unknown formats
        with open(self.metadata_path, 'w') as f:
            f.write('cache-formats = foo bar\n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.cache_format == 'md5-dict'
        assert 'unknown cache format:' in caplog.text
        caplog.clear()
        del repo_config

        # known format
        with open(self.metadata_path, 'w') as f:
            f.write('cache-formats = pms\n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.cache_format == 'pms'
        del repo_config

        # multiple formats -- favored format is selected
        with open(self.metadata_path, 'w') as f:
            f.write('cache-formats = pms md5-dict\n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.cache_format == 'md5-dict'
        del repo_config

        # unknown + known
        with open(self.metadata_path, 'w') as f:
            f.write('cache-formats = foo md5-dict\n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.cache_format == 'md5-dict'
        del repo_config

    def test_profile_formats(self, caplog):
        os.mkdir(self.profiles_base)
        with open(os.path.join(self.profiles_base, 'repo_name'), 'w') as f:
            f.write('pms_name')

        # empty repo
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.profile_formats == {'pms'}
        del repo_config
        caplog.clear()

        # explicit empty setting
        os.mkdir(os.path.dirname(self.metadata_path))
        with open(self.metadata_path, 'w') as f:
            f.write('masters =\nprofile-formats =\n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.profile_formats == {'pms'}
        assert not caplog.text
        caplog.clear()
        del repo_config
        # message shown at info log level
        caplog.set_level(logging.INFO)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert 'has explicitly unset profile-formats' in caplog.text
        caplog.clear()
        del repo_config

        # unknown formats
        caplog.set_level(logging.WARNING)
        with open(self.metadata_path, 'w') as f:
            f.write('masters =\nprofile-formats = foo bar\n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.profile_formats == {'pms'}
        assert not caplog.text
        caplog.clear()
        del repo_config
        # message shown at info log level
        caplog.set_level(logging.INFO)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert 'has unsupported profile format' in caplog.text
        caplog.clear()
        del repo_config

        # unknown + known
        caplog.set_level(logging.WARNING)
        with open(self.metadata_path, 'w') as f:
            f.write('masters =\nprofile-formats = foo portage-2\n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.profile_formats == {'pms', 'portage-2'}
        assert not caplog.text
        caplog.clear()
        del repo_config
        # message shown at info log level
        caplog.set_level(logging.INFO)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert 'has unsupported profile format' in caplog.text
        caplog.clear()
        del repo_config

        # known formats
        caplog.set_level(logging.WARNING)
        with open(self.metadata_path, 'w') as f:
            f.write('profile-formats = portage-1 portage-2\n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.profile_formats == {'portage-1', 'portage-2'}
        del repo_config

    def test_pms_repo_name(self):
        os.mkdir(self.profiles_base)
        repo_name_path = os.path.join(self.profiles_base, 'repo_name')

        # nonexistent file
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.pms_repo_name is None
        del repo_config

        # empty file
        touch(repo_name_path)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.pms_repo_name == ''
        del repo_config

        # whitespace
        with open(repo_name_path, 'w') as f:
            f.write(' \n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.pms_repo_name == ''
        del repo_config

        # whitespace + name
        with open(repo_name_path, 'w') as f:
            f.write(' repo \n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.pms_repo_name == 'repo'
        del repo_config

        # regular name
        with open(repo_name_path, 'w') as f:
            f.write('newrepo')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.pms_repo_name == 'newrepo'
        del repo_config

        # regular name EOLed
        with open(repo_name_path, 'w') as f:
            f.write('newrepo2\n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.pms_repo_name == 'newrepo2'
        del repo_config

        # multi-line
        with open(repo_name_path, 'w') as f:
            f.write('newrepo3\nfoobar')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.pms_repo_name == 'newrepo3'
        del repo_config

        # binary data
        with open(repo_name_path, 'wb') as f:
            f.write(b'\x6e\x65\x77\x72\x65\x70\x6f\x34')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.pms_repo_name == 'newrepo4'
        del repo_config

    def test_repo_id(self, caplog):
        # nonexistent repo
        repo_config = repo_objs.RepoConfig('nonexistent')
        assert repo_config.repo_id == 'nonexistent'
        del repo_config

        # empty repo
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.repo_id == self.repo_path
        assert caplog.text == ''
        caplog.clear()
        del repo_config

        # nonempty repo
        os.mkdir(self.profiles_base)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.repo_id == self.repo_path
        assert 'repo lacks a defined name:' in caplog.text
        caplog.clear()
        del repo_config

        # pms repo name exists
        with open(os.path.join(self.profiles_base, 'repo_name'), 'w') as f:
            f.write('pms_name')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.repo_id == 'pms_name'
        del repo_config

        # layout.conf repo name exists
        os.mkdir(os.path.dirname(self.metadata_path))
        with open(self.metadata_path, 'w') as f:
            f.write('repo-name = metadata_name')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.repo_id == 'metadata_name'
        del repo_config

        # config name exists
        repo_config = repo_objs.RepoConfig(self.repo_path, config_name='config_name')
        assert repo_config.repo_id == 'config_name'
        del repo_config

    def test_known_arches(self):
        # nonexistent repo
        repo_config = repo_objs.RepoConfig('nonexistent')
        assert repo_config.known_arches == frozenset()
        del repo_config

        # empty file
        os.mkdir(self.profiles_base)
        arches_path = os.path.join(self.profiles_base, 'arch.list')
        touch(arches_path)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.known_arches == frozenset()
        del repo_config

        # single entry
        with open(arches_path, 'w') as f:
            f.write('foo')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.known_arches == frozenset(['foo'])
        del repo_config

        # multiple entries with whitespaces and comments
        with open(arches_path, 'w') as f:
            f.write(
                """
                amd64
                x86

                # prefix
                foo-bar
                """)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.known_arches == frozenset(['amd64', 'x86', 'foo-bar'])
        del repo_config

    def test_arches_desc(self):
        # nonexistent repo
        repo_config = repo_objs.RepoConfig('nonexistent')
        empty = {'stable': set(), 'transitional': set(), 'testing': set()}
        assert repo_config.arches_desc == ImmutableDict(empty)
        del repo_config

        # empty file
        os.mkdir(self.profiles_base)
        arches_desc_path = os.path.join(self.profiles_base, 'arches.desc')
        touch(arches_desc_path)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.arches_desc == ImmutableDict(empty)
        del repo_config

        # regular entries
        with open(os.path.join(self.profiles_base, 'arch.list'), 'w') as f:
            f.write(
                """
                amd64
                alpha
                foo
                """)
        with open(arches_desc_path, 'w') as f:
            f.write(
                """
                # arches.desc file

                amd64 stable
                alpha testing
                """)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.arches_desc['stable'] == {'amd64'}
        assert repo_config.arches_desc['testing'] == {'alpha'}
        assert repo_config.arches_desc['transitional'] == set()
        del repo_config

    def test_use_desc(self):
        # nonexistent repo
        repo_config = repo_objs.RepoConfig('nonexistent')
        assert repo_config.use_desc == ()
        del repo_config

        # empty file
        os.mkdir(self.profiles_base)
        use_desc_path = os.path.join(self.profiles_base, 'use.desc')
        touch(use_desc_path)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.use_desc == ()
        del repo_config

        # regular entries
        with open(use_desc_path, 'w') as f:
            f.write(
                """
                # copy
                # license

                foo1 - enable foo1
                foo2 - enable foo2
                bar3 - add bar3 support
                """)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert 3 == len(repo_config.use_desc)
        del repo_config

    def test_use_expand_desc(self):
        # nonexistent repo
        repo_config = repo_objs.RepoConfig('nonexistent')
        assert repo_config.use_expand_desc == {}
        del repo_config

        # empty file
        use_expand_desc_path = os.path.join(self.profiles_base, 'desc')
        os.makedirs(use_expand_desc_path)
        use_expand_desc_file = os.path.join(use_expand_desc_path, 'foo.desc')
        touch(use_expand_desc_file)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.use_expand_desc == {'foo': ()}
        del repo_config

        # regular entries
        with open(use_expand_desc_file, 'w') as f:
            f.write(
                """
                # copy
                # license

                bar - add bar support
                baz - build using baz
                """)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.use_expand_desc == {
            'foo': (
                ('foo_bar', 'add bar support'),
                ('foo_baz', 'build using baz')
            )}
        del repo_config

    def test_use_local_desc(self):
        # nonexistent repo
        repo_config = repo_objs.RepoConfig('nonexistent')
        assert repo_config.use_local_desc == ()
        del repo_config

        # empty file
        os.mkdir(self.profiles_base)
        use_local_desc_path = os.path.join(self.profiles_base, 'use.local.desc')
        touch(use_local_desc_path)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.use_local_desc == ()
        del repo_config

        # regular entries
        with open(use_local_desc_path, 'w') as f:
            f.write(
                """
                # copy
                # license

                cat/pkg1:foo1 - enable foo1
                cat1/pkg2:foo2 - enable foo2
                cat2/pkg3:bar3 - add bar3 support
                """)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert 3 == len(repo_config.use_local_desc)
        del repo_config

    def test_updates(self):
        # nonexistent repo
        repo_config = repo_objs.RepoConfig('nonexistent')
        assert repo_config.updates == {}
        del repo_config

        # empty file
        updates_path = os.path.join(self.profiles_base, 'updates')
        updates_file_path = os.path.join(updates_path, '1Q-2019')
        os.makedirs(updates_path)
        touch(updates_file_path)
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.updates == {}
        del repo_config

        # simple pkg move
        # TODO: move pkg_updates content tests to its own module
        with open(updates_file_path, 'w') as f:
            f.write('move cat1/pkg1 cat2/pkg1\n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        expected_updates = {
            'cat1/pkg1':  [('move', atom.atom('cat1/pkg1'), atom.atom('cat2/pkg1'))],
        }
        assert repo_config.updates == expected_updates
        del repo_config

        # extraneous file should be ignored
        extra_file_path = os.path.join(updates_path, 'frobnicate')
        with open(extra_file_path, 'w') as f:
            f.write('move cat1/pkg2 cat1/pkg3\n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.updates == expected_updates
        del repo_config

    def test_updates_eapi8(self):
        # empty file
        updates_path = os.path.join(self.profiles_base, 'updates')
        updates_file_path = os.path.join(updates_path, '2021.1')
        os.makedirs(updates_path)
        touch(updates_file_path)
        with open(os.path.join(self.profiles_base, 'eapi'), 'w') as f:
          f.write('8\n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.updates == {}
        del repo_config

        # simple pkg move
        # TODO: move pkg_updates content tests to its own module
        with open(updates_file_path, 'w') as f:
            f.write('move cat1/pkg1 cat2/pkg1\n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        expected_updates = {
            'cat1/pkg1':  [('move', atom.atom('cat1/pkg1'), atom.atom('cat2/pkg1'))],
        }
        assert repo_config.updates == expected_updates
        del repo_config

        # extraneous file should be ignored
        extra_file_path = os.path.join(updates_path, '.frobnicate')
        with open(extra_file_path, 'w') as f:
            f.write('move cat1/pkg2 cat1/pkg3\n')
        repo_config = repo_objs.RepoConfig(self.repo_path)
        assert repo_config.updates == expected_updates
        del repo_config



__all__ = ("nodeps_repo", "caching_repo")

from snakeoil.iterables import caching_iter, iter_sort
from snakeoil.klass import DirProxy, GetAttrProxy

from ..ebuild.conditionals import DepSet
from ..operations.repo import operations_proxy
from ..package.mutated import MutatedPkg
from ..restrictions import packages


class nodeps_repo:
    """Repository wrapper that returns wrapped pkgs with deps wiped."""

    default_bdepend = default_depend = default_rdepend = default_pdepend = default_idepend = DepSet()

    def __init__(self, repo):
        """
        :param repo: repository to wrap
        """
        self.raw_repo = repo

    def itermatch(self, *a, **kwds):
        return (MutatedPkg(
            x, overrides={"bdepend": self.default_bdepend,
                          "depend": self.default_depend,
                          "rdepend": self.default_rdepend,
                          "pdepend": self.default_pdepend,
                          "idepend": self.default_idepend}
            )
            for x in self.raw_repo.itermatch(*a, **kwds))

    def match(self, *a, **kwds):
        return list(self.itermatch(*a, **kwds))

    __getattr__ = GetAttrProxy("raw_repo")
    __dir__ = DirProxy("raw_repo")

    def __iter__(self):
        return self.itermatch(packages.AlwaysTrue)


class restrict_repo:
    """Repository wrapper that skips packages matching a given restriction."""

    def __init__(self, restrict, repo):
        """
        :param restrict: package matching restriction
        :param repo: repository to wrap
        """
        self.raw_repo = repo
        self.restrict = restrict

    def itermatch(self, *a, **kwds):
        return (
            x for x in self.raw_repo.itermatch(*a, **kwds)
            if not self.restrict.match(x))

    def match(self, *a, **kwds):
        return list(self.itermatch(*a, **kwds))

    __getattr__ = GetAttrProxy("raw_repo")
    __dir__ = DirProxy("raw_repo")

    def __iter__(self):
        return self.itermatch(packages.AlwaysTrue)


class caching_repo:
    """Repository wrapper that overrides match, returning :obj:`caching_iter` instances.

    Itermatch is slaved to match, in other words iterating over the
    caching_iter.

    Main use for this is to cache results from query lookups;
    if matches restrict arg is in the cache, the caller gets a shared
    caching_iter sequence, which may already be fully loaded with pkg
    instances.

    This can boost random lookup time pretty nicely, while helping to
    hold instance in memory to avoid redoing work.

    Cost of this of course is that involved objects are forced to stay
    in memory till the cache is cleared.  General use, not usually what
    you want- if you're making a lot of random queries that are duplicates
    (resolver does this for example), caching helps.
    """

    operations_kls = operations_proxy

    def __init__(self, db, strategy):
        """
        :param db: an instance supporting the repository protocol to cache
          queries from.
        :param strategy: forced sorting strategy for results.  If you don't
          need sorting, pass in iter.
        """
        self.__db__ = db
        self.__strategy__ = strategy
        self.__cache__ = {}

    def match(self, restrict):
        v = self.__cache__.get(restrict)
        if v is None:
            v = self.__cache__[restrict] = \
                caching_iter(
                    self.__db__.itermatch(restrict, sorter=self.__strategy__))
        return v

    def itermatch(self, restrict):
        return iter(self.match(restrict))

    __getattr__ = GetAttrProxy("__db__")
    __dir__ = DirProxy("__db__")

    def clear(self):
        self.__cache__.clear()


class multiplex_sorting_repo:

    def __init__(self, sorter, repos):
        self.__repos__ = tuple(repos)
        self.__sorter__ = sorter

    def itermatch(self, restrict):
        repo_iters = [repo.itermatch(restrict) for repo in self.__repos__]
        return iter_sort(self.__sorter__, *repo_iters)

    def match(self, restrict):
        return list(self.itermatch(restrict))

    def has_match(self, restrict):
        for repo in self.__repos__:
            if repo.has_match(restrict):
                return True
        return False



import os

from pkgcore.fs import fs
from snakeoil.chksum import get_chksums
from snakeoil.data_source import data_source
from snakeoil.osutils import normpath, pjoin
from snakeoil.test import TestCase
from snakeoil.test.mixins import tempdir_decorator


class base:

    kls = None

    def make_obj(self, location="/tmp/foo", **kwds):
        kwds.setdefault("strict", False)
        return self.kls(location, **kwds)

    def test_basename(self):
        self.assertEqual(self.make_obj(location='/asdf').basename, 'asdf')
        self.assertEqual(self.make_obj(location='/a/b').basename, 'b')

    def test_dirname(self):
        self.assertEqual(self.make_obj(location='/asdf').dirname, '/')
        self.assertEqual(self.make_obj(location='/a/b').dirname, '/a')

    def test_location_normalization(self):
        for loc in ('/tmp/a', '/tmp//a', '/tmp//', '/tmp/a/..'):
            self.assertEqual(self.make_obj(location=loc).location,
                normpath(loc), reflective=False)

    def test_change_attributes(self):
        # simple test...
        o = self.make_obj("/foon")
        self.assertNotEqual(o, o.change_attributes(location="/nanners"))

    def test_init(self):
        mkobj = self.make_obj
        o = mkobj("/tmp/foo")
        self.assertEqual(o.location, "/tmp/foo")
        self.assertEqual(mkobj(mtime=100).mtime, 100)
        self.assertEqual(mkobj(mode=0o660).mode, 0o660)
        # ensure the highband stays in..
        self.assertEqual(mkobj(mode=0o42660).mode, 0o42660)
        self.assertEqual(mkobj(uid=0).uid, 0)
        self.assertEqual(mkobj(gid=0).gid, 0)

    def test_hash(self):
        # might seem odd, but done this way to avoid the any potential
        # false positives from str's hash returning the same
        d = {self.make_obj("/tmp/foo"):None}
        # ensure it's accessible without a KeyError
        d[self.make_obj("/tmp/foo")]

    def test_eq(self):
        o = self.make_obj("/tmp/foo")
        self.assertEqual(o, self.make_obj("/tmp/foo"))
        self.assertNotEqual(o, self.make_obj("/tmp/foo2"))

    def test_setattr(self):
        o = self.make_obj()
        for attr in o.__attrs__:
            self.assertRaises(AttributeError, setattr, o, attr, "monkies")

    @tempdir_decorator
    def test_realpath(self):
        # just to be safe, since this could trash some tests.
        self.dir = os.path.realpath(self.dir)
        os.mkdir(pjoin(self.dir, "test1"))
        obj = self.make_obj(location=pjoin(self.dir, "test1", "foon"))
        self.assertIdentical(obj, obj.realpath())
        os.symlink(pjoin(self.dir, "test1"), pjoin(self.dir, "test2"))
        obj = self.make_obj(location=pjoin(self.dir, "test2", "foon"))
        new_obj = obj.realpath()
        self.assertNotIdentical(obj, new_obj)
        self.assertEqual(new_obj.location, pjoin(self.dir, "test1", "foon"), reflective=False)
        os.symlink(pjoin(self.dir, "test3"), pjoin(self.dir, "nonexistent"))
        obj = self.make_obj(pjoin(self.dir, "nonexistent", "foon"))
        # path is incomplete; should still realpath it.
        new_obj = obj.realpath()
        self.assertNotIdentical(obj, new_obj)
        self.assertEqual(new_obj.location, pjoin(self.dir, "test3", "foon"))

    def test_default_attrs(self):
        self.assertEqual(self.make_obj(location="/adsf").mode, None)
        class tmp(self.kls):
            __default_attrs__ = self.kls.__default_attrs__.copy()
            __default_attrs__['tmp'] = lambda self2:getattr(self2, 'a', 1)
            __attrs__ = self.kls.__attrs__ + ('tmp',)
            __slots__ = ('a', 'tmp')
        try:
            self.kls = tmp
            self.assertEqual(self.make_obj('/adsf', strict=False).tmp, 1)
            t = self.make_obj('/asdf', a='foon', strict=False)
            self.assertEqual(t.tmp, "foon")
        finally:
            del self.kls


class Test_fsFile(TestCase, base):

    kls = fs.fsFile

    def test_init(self):
        base.test_init(self)
        mkobj = self.make_obj
        o = mkobj(__file__)
        with open(__file__) as f:
            raw_data = f.read()
        self.assertEqual(o.data.text_fileobj().read(), raw_data)
        o = mkobj("/bin/this-file-should-not-exist-nor-be-read",
            data=data_source(raw_data))
        self.assertEqual(o.data.text_fileobj().read(), raw_data)
        keys = list(o.chksums.keys())
        self.assertEqual([o.chksums[x] for x in keys],
            list(get_chksums(data_source(raw_data), *keys)))

        chksums = dict(iter(o.chksums.items()))
        self.assertEqual(sorted(mkobj(chksums=chksums).chksums.items()),
            sorted(chksums.items()))

    def test_chksum_regen(self):
        data_source = object()
        obj = self.make_obj(__file__)
        self.assertIdentical(obj.chksums,
            obj.change_attributes(location="/tpp").chksums)
        chksums1 = obj.chksums
        self.assertNotIdentical(chksums1,
            obj.change_attributes(data=data_source).chksums)

        self.assertIdentical(chksums1,
            obj.change_attributes(data=data_source,
                chksums=obj.chksums).chksums)

        obj2 = self.make_obj(__file__, chksums={1:2})
        self.assertIdentical(obj2.chksums,
            obj2.change_attributes(data=data_source).chksums)


class Test_fsLink(TestCase, base):
    kls = fs.fsLink

    def make_obj(self, location="/tmp/foo", **kwds):
        target = kwds.pop("target", pjoin(location, "target"))
        kwds.setdefault("strict", False)
        return self.kls(location, target, **kwds)

    def test_init(self):
        base.test_init(self)
        mkobj = self.make_obj
        self.assertEqual(mkobj(target="k9").target, "k9")
        self.assertEqual(mkobj(target="../foon").target, "../foon")

    def test_resolved_target(self):
        self.assertEqual(self.make_obj(location="/tmp/foon", target="dar").resolved_target,
            "/tmp/dar")
        self.assertEqual(self.make_obj(location="/tmp/foon", target="/dar").resolved_target,
            "/dar")

    def test_cmp(self):
        obj1 = self.make_obj(
            location='/usr/lib64/opengl/nvidia/lib/libnvidia-tls.so.1',
            target='../tls/libnvidia-tls.so.1')
        obj2 = self.make_obj(
            location='/usr/lib32/opengl/nvidia/lib/libGL.s',
            target='libGL.so.173.14.09')
        self.assertTrue(obj1 > obj2)
        self.assertTrue(obj2 < obj1)


class Test_fsDev(TestCase, base):
    kls = fs.fsDev

    def test_init(self):
        base.test_init(self)
        mkobj = self.make_obj
        self.assertRaises(TypeError, mkobj, major=-1, strict=True)
        self.assertRaises(TypeError, mkobj, minor=-1, strict=True)
        self.assertEqual(mkobj(major=1).major, 1)
        self.assertEqual(mkobj(minor=1).minor, 1)


class Test_fsFifo(TestCase, base):
    kls = fs.fsFifo


class Test_fsDir(TestCase, base):
    kls = fs.fsDir


class Test_Modules_Funcs(TestCase):

    def test_is_funcs(self):
        # verify it intercepts the missing attr
        self.assertFalse(fs.isdir(object()))
        self.assertFalse(fs.isreg(object()))
        self.assertFalse(fs.isfifo(object()))

        self.assertTrue(fs.isdir(fs.fsDir('/tmp', strict=False)))
        self.assertFalse(fs.isreg(fs.fsDir('/tmp', strict=False)))
        self.assertTrue(fs.isreg(fs.fsFile('/tmp', strict=False)))



from pkgcore.scripts import pclean
from pkgcore.test.scripts.helpers import ArgParseMixin
from snakeoil.test import TestCase


class CommandlineTest(TestCase, ArgParseMixin):

    _argparser = pclean.argparser

    suppress_domain = True

    def test_parser(self):
        self.assertError('the following arguments are required: subcommand')



#!/usr/bin/env python3

import itertools
import sys

try:
    from pkgcore.restrictions.boolean import OrRestriction
    from pkgcore.util import commandline
except ImportError:
    print('Cannot import pkgcore!', file=sys.stderr)
    print('Verify it is properly installed and/or PYTHONPATH is set correctly.', file=sys.stderr)
    if '--debug' not in sys.argv:
        print('Add --debug to the commandline for a traceback.', file=sys.stderr)
    else:
        raise
    sys.exit(1)


argparser = commandline.ArgumentParser(color=False, version=False)
argparser.add_argument(
    'target', nargs='+', help='target package atoms')


@argparser.bind_final_check
def check_args(parser, namespace):
    namespace.repo = namespace.domain.ebuild_repos
    namespace.restrict = OrRestriction(
        *commandline.convert_to_restrict(namespace.target))


def getter(pkg):
    return (pkg.key, getattr(pkg, "maintainers", None))


@argparser.bind_main_func
def main(options, out, err):
    for t, pkgs in itertools.groupby(
            options.repo.itermatch(options.restrict, sorter=sorted), getter):
        out.write(t[0])
        out.first_prefix = "    "
        for pkg in pkgs:
            out.write('%s::%s' % (pkg.cpvstr, pkg.repo.repo_id))
        out.first_prefix = ""
        item = 'maintainer'
        values = t[1]
        if values:
            out.write(
                "%s%s: %s" %
                (item.title(), 's'[len(values) == 1:], ', '.join(str(x) for x in values)))
        out.write()


if __name__ == '__main__':
    tool = commandline.Tool(argparser)
    sys.exit(tool())



#!/usr/bin/env python3

import glob
import os
import subprocess
import sys
from itertools import chain

from setuptools import setup
from setuptools._distutils import log
from setuptools._distutils.errors import DistutilsExecError
from setuptools._distutils.util import byte_compile
from snakeoil.dist import distutils_extensions as pkgdist

pkgdist_setup, pkgdist_cmds = pkgdist.setup()

# These offsets control where we install the pkgcore config files and the EBD
# bits relative to the install-data path given to the install subcmd.
DATA_INSTALL_OFFSET = 'share/pkgcore'
CONFIG_INSTALL_OFFSET = os.path.join(DATA_INSTALL_OFFSET, 'config')
LIBDIR_INSTALL_OFFSET = 'lib/pkgcore'
EBD_INSTALL_OFFSET = os.path.join(LIBDIR_INSTALL_OFFSET, 'ebd')


class sdist(pkgdist.sdist):
    """sdist wrapper to bundle generated files for release."""

    def make_release_tree(self, base_dir, files):
        """Generate bash function lists for releases."""
        import shutil

        # generate function lists so they don't need to be created on install
        write_pkgcore_ebd_funclists(root='/', target='ebd/.generated')
        write_pkgcore_ebd_cmdlists(root='/', target='ebd/.generated')
        write_pkgcore_ebd_eapi_libs(root='/', target='ebd/.generated')
        shutil.copytree(
            os.path.join(pkgdist.REPODIR, 'ebd', '.generated'),
            os.path.join(base_dir, 'ebd', '.generated'))

        pkgdist.sdist.make_release_tree(self, base_dir, files)


class install(pkgdist.install):
    """Install wrapper to generate and install pkgcore-related files."""

    def run(self):
        super().run()
        target = self.install_data
        root = self.root or '/'
        if target.startswith(root):
            target = os.path.join('/', os.path.relpath(target, root))
        target = os.path.abspath(target)
        if not self.dry_run:
            # Install module plugincache
            # TODO: move this to pkgdist once plugin support is moved to snakeoil
            with pkgdist.syspath(pkgdist.PACKAGEDIR):
                from pkgcore import plugin, plugins
                log.info('Generating plugin cache')
                path = os.path.join(self.install_purelib, 'pkgcore', 'plugins')
                plugin.initialize_cache(plugins, force=True, cache_dir=path)

            # Install configuration data so pkgcore knows where to find its content,
            # rather than assuming it is running from a tarball/git repo.
            write_pkgcore_lookup_configs(self.install_purelib, target)

            # Generate ebd libs when not running from release tarballs that
            # contain pre-generated files.
            if not os.path.exists(os.path.join(pkgdist.REPODIR, 'man')):
                generated_target = os.path.join(target, EBD_INSTALL_OFFSET, '.generated')
                write_pkgcore_ebd_funclists(root=root, target=generated_target)
                write_pkgcore_ebd_cmdlists(root=root, target=generated_target)
                write_pkgcore_ebd_eapi_libs(root=root, target=generated_target)


def write_pkgcore_ebd_funclists(root, target):
    "Generate bash function lists from ebd implementation for env filtering."""
    ebd_dir = target
    if root != '/':
        ebd_dir = os.path.join(root, target.lstrip('/'))
    os.makedirs(os.path.join(ebd_dir, 'funcs'), exist_ok=True)

    # generate global function list
    path = os.path.join(ebd_dir, 'funcs', 'global')
    log.info(f'writing ebd global function list: {path!r}')
    with open(path, 'w') as f:
        if subprocess.call(
                [os.path.join(pkgdist.REPODIR, 'ebd', 'generate_global_func_list')],
                cwd=ebd_dir, stdout=f):
            raise DistutilsExecError("generating global function list failed")

    # generate EAPI specific function lists
    with pkgdist.syspath(pkgdist.PACKAGEDIR):
        from pkgcore.ebuild.eapi import EAPI
        for eapi_obj in EAPI.known_eapis.values():
            eapi = str(eapi_obj)
            path = os.path.join(ebd_dir, 'funcs', eapi)
            log.info(f'writing EAPI {eapi} function list: {path!r}')
            with open(path, 'w') as f:
                if subprocess.call(
                        [os.path.join(pkgdist.REPODIR, 'ebd', 'generate_eapi_func_list'), eapi],
                        cwd=ebd_dir, stdout=f):
                    raise DistutilsExecError(
                        "generating EAPI %s function list failed" % eapi)


def write_pkgcore_ebd_cmdlists(root, target):
    "Generate bash function lists from ebd implementation for env filtering."""
    ebd_dir = target
    if root != '/':
        ebd_dir = os.path.join(root, target.lstrip('/'))
    os.makedirs(os.path.join(ebd_dir, 'cmds'), exist_ok=True)

    # generate EAPI specific command lists
    with pkgdist.syspath(pkgdist.PACKAGEDIR):
        from pkgcore.ebuild.eapi import EAPI
        for eapi_obj in EAPI.known_eapis.values():
            eapi = str(eapi_obj)
            os.makedirs(os.path.join(ebd_dir, 'cmds', eapi), exist_ok=True)

            path = os.path.join(ebd_dir, 'cmds', eapi, 'banned')
            log.info(f'writing EAPI {eapi} banned command list: {path!r}')
            with open(path, 'w') as f:
                if subprocess.call(
                        [os.path.join(pkgdist.REPODIR, 'ebd', 'generate_eapi_cmd_list'), '-b', eapi],
                        cwd=ebd_dir, stdout=f):
                    raise DistutilsExecError(f'generating EAPI {eapi} banned command list failed')

            path = os.path.join(ebd_dir, 'cmds', eapi, 'deprecated')
            log.info(f'writing EAPI {eapi} deprecated command list: {path!r}')
            with open(path, 'w') as f:
                if subprocess.call(
                        [os.path.join(pkgdist.REPODIR, 'ebd', 'generate_eapi_cmd_list'), '-d', eapi],
                        cwd=ebd_dir, stdout=f):
                    raise DistutilsExecError(f'generating EAPI {eapi} deprecated command list failed')

            path = os.path.join(ebd_dir, 'cmds', eapi, 'internal')
            log.info(f'writing EAPI {eapi} internal command list: {path!r}')
            with open(path, 'w') as f:
                if subprocess.call(
                        [os.path.join(pkgdist.REPODIR, 'ebd', 'generate_eapi_cmd_list'), '-i', eapi],
                        cwd=ebd_dir, stdout=f):
                    raise DistutilsExecError(f'generating EAPI {eapi} internal command list failed')


def write_pkgcore_ebd_eapi_libs(root, target):
    "Generate bash EAPI scope libs for sourcing."""
    ebd_dir = target
    if root != '/':
        ebd_dir = os.path.join(root, target.lstrip('/'))

    script = os.path.join(pkgdist.REPODIR, 'ebd', 'generate_eapi_lib')
    with pkgdist.syspath(pkgdist.PACKAGEDIR):
        from pkgcore.ebuild.eapi import EAPI
        for eapi_obj in EAPI.known_eapis.values():
            eapi = str(eapi_obj)
            os.makedirs(os.path.join(ebd_dir, 'libs', eapi), exist_ok=True)

            # generate global scope lib
            path = os.path.join(ebd_dir, 'libs', eapi, 'global')
            log.info(f'writing global EAPI {eapi} lib: {path!r}')
            with open(path, 'w') as f:
                if subprocess.call([script, eapi], cwd=ebd_dir, stdout=f):
                    raise DistutilsExecError(
                        f"generating global scope EAPI {eapi} lib failed")

            for phase in eapi_obj.phases.values():
                # generate phase scope lib
                path = os.path.join(ebd_dir, 'libs', eapi, phase)
                log.info(f'writing EAPI {eapi} {phase} phase lib: {path!r}')
                with open(path, 'w') as f:
                    if subprocess.call([script, '-s', phase, eapi], cwd=ebd_dir, stdout=f):
                        raise DistutilsExecError(
                            f"generating {phase} phase scope EAPI {eapi} lib failed")


def write_pkgcore_lookup_configs(python_base, install_prefix, injected_bin_path=()):
    """Generate file of install path constants."""
    path = os.path.join(python_base, "pkgcore", "_const.py")
    os.makedirs(os.path.dirname(path), exist_ok=True)
    log.info("writing lookup config to %r" % path)

    wheel_install = (
        install_prefix != os.path.abspath(sys.prefix)
        and not install_prefix.startswith(pkgdist.REPODIR)
    )

    with open(path, "w") as f:
        os.chmod(path, 0o644)
        # write more dynamic _const file for wheel installs
        if wheel_install:
            import textwrap
            f.write(textwrap.dedent("""\
                import os.path as osp
                import sys

                from snakeoil import process

                INSTALL_PREFIX = osp.abspath(sys.prefix)
                DATA_PATH = osp.join(INSTALL_PREFIX, {!r})
                CONFIG_PATH = osp.join(INSTALL_PREFIX, {!r})
                LIBDIR_PATH = osp.join(INSTALL_PREFIX, {!r})
                EBD_PATH = osp.join(INSTALL_PREFIX, {!r})
                INJECTED_BIN_PATH = ()
                CP_BINARY = process.find_binary('cp')
            """.format(
                DATA_INSTALL_OFFSET, CONFIG_INSTALL_OFFSET,
                LIBDIR_INSTALL_OFFSET, EBD_INSTALL_OFFSET)))
        else:
            f.write("INSTALL_PREFIX=%r\n" % install_prefix)
            f.write("DATA_PATH=%r\n" %
                    os.path.join(install_prefix, DATA_INSTALL_OFFSET))
            f.write("CONFIG_PATH=%r\n" %
                    os.path.join(install_prefix, CONFIG_INSTALL_OFFSET))
            f.write("LIBDIR_PATH=%r\n" %
                    os.path.join(install_prefix, LIBDIR_INSTALL_OFFSET))
            f.write("EBD_PATH=%r\n" %
                    os.path.join(install_prefix, EBD_INSTALL_OFFSET))

            # This is added to suppress the default behaviour of looking
            # within the repo for a bin subdir.
            f.write("INJECTED_BIN_PATH=%r\n" % (tuple(injected_bin_path),))

            # Static paths for various utilities.
            from snakeoil import process
            required_progs = ('cp',)
            try:
                for prog in required_progs:
                    prog_path = process.find_binary(prog)
                    f.write("%s_BINARY=%r\n" % (prog.upper(), prog_path))
            except process.CommandNotFound:
                raise DistutilsExecError(
                    "generating lookup config failed: required utility %r missing from PATH" % (prog,))

            f.close()
            byte_compile([path], prefix=python_base)
            byte_compile([path], optimize=1, prefix=python_base)
            byte_compile([path], optimize=2, prefix=python_base)


setup(**dict(
    pkgdist_setup,
    description='package managing framework',
    url='https://github.com/pkgcore/pkgcore',
    license='BSD',
    author='Tim Harder',
    author_email='radhermit@gmail.com',
    entry_points={'pytest11': ['pkgcore = pkgcore.pytest.plugin']},
    data_files=list(chain(
        pkgdist.data_mapping(EBD_INSTALL_OFFSET, 'ebd'),
        pkgdist.data_mapping(DATA_INSTALL_OFFSET, 'data'),
        pkgdist.data_mapping('share/zsh/site-functions', 'shell/zsh/completion'),
        pkgdist.data_mapping(
            os.path.join(LIBDIR_INSTALL_OFFSET, 'shell'), 'shell',
            skip=glob.glob('shell/*/completion'),
        ),
    )),
    cmdclass=dict(
        pkgdist_cmds,
        sdist=sdist,
        install=install,
    ),
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
))



"""base package class; instances should derive from this.

Right now, doesn't provide much, need to change that down the line
"""

__all__ = ("base", "wrapper", "dynamic_getattr_dict", "DynamicGetattrSetter")

import itertools

from snakeoil import klass, sequences

from .. import exceptions as base_errors
from ..operations import format
from . import errors


class base(klass.SlotsPicklingMixin, metaclass=klass.immutable_instance):

    built = False
    configurable = False
    _operations = format.operations

    __slots__ = ("__weakref__",)
    _get_attr = {}

    @property
    def versioned_atom(self):
        raise NotImplementedError(self, "versioned_atom")

    @property
    def unversioned_atom(self):
        raise NotImplementedError(self, "unversioned_atom")

    def operations(self, domain, **kwds):
        return self._operations(domain, self, **kwds)

    @property
    def is_supported(self):
        return True


class wrapper(base):

    __slots__ = ("_raw_pkg", "_domain")

    def operations(self, domain, **kwds):
        return self._raw_pkg._operations(domain, self, **kwds)

    def __init__(self, raw_pkg):
        object.__setattr__(self, "_raw_pkg", raw_pkg)

    def __eq__(self, other):
        if isinstance(other, wrapper):
            return self._raw_pkg == other._raw_pkg
        try:
            return self._raw_pkg == other
        except TypeError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if isinstance(other, wrapper):
            return self._raw_pkg < other._raw_pkg
        return self._raw_pkg < other

    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)

    def __gt__(self, other):
        if isinstance(other, wrapper):
            return self._raw_pkg > other._raw_pkg
        return self._raw_pkg > other

    def __ge__(self, other):
        return self.__gt__(other) or self.__eq__(other)

    __getattr__ = klass.GetAttrProxy("_raw_pkg")
    __dir__ = klass.DirProxy("_raw_pkg")

    _get_attr = klass.alias_attr("_raw_pkg._get_attr")
    built = klass.alias_attr("_raw_pkg.built")
    versioned_atom = klass.alias_attr("_raw_pkg.versioned_atom")
    unversioned_atom = klass.alias_attr("_raw_pkg.unversioned_atom")
    is_supported = klass.alias_attr('_raw_pkg.is_supported')

    def __hash__(self):
        return hash(self._raw_pkg)


def dynamic_getattr_dict(self, attr):
    functor = self._get_attr.get(attr)
    if functor is None:
        raise AttributeError(self, attr)
    try:
        val = functor(self)
        object.__setattr__(self, attr, val)
        return val
    except errors.MetadataException as e:
        if e.attr == attr:
            raise
        raise errors.MetadataException(self, attr, e.error, e.verbose) from e
    except (errors.PackageError, UnicodeDecodeError) as e:
        raise errors.MetadataException(self, attr, str(e)) from e
    except PermissionError as e:
        raise base_errors.PermissionDenied(self.path, write=False) from e


class DynamicGetattrSetter(type):
    """Metaclass utilizing __getattr__ to JIT generate attributes and store them.

    Consider `snakeoil.klass.jit_attr` for comparison; that pseudo property
    will invoke a functor and store the result, but every subsequent access- still
    pays overhead of passing through the redirects.

    This metaclass lacks that overhead; via hooking __getattr__, this generates
    the requested attribute, stores it on the instance, and returns it.  All
    future access of that attribute go through the fast path cpy access ways.

    This optimization in the early days of pkgcore had drastic impact; in modern
    times python has improved.  There still is gain, but the implementation complexity
    may warrant phasing this out in favor of `snakeoil.klas.jit_attr` alternatives.
    """

    class register:
        """Decorator used to mark a function as an attribute loader."""

        __slots__ = ('functor',)

        def __init__(self, functor):
            self.functor = functor

    def __new__(cls, name, bases, class_dict):
        new_functions = {
            attr: class_dict.pop(attr).functor
            for attr, thing in list(class_dict.items())
            if isinstance(thing, cls.register)
        }

        existing = {}
        for base in bases:
            existing.update(getattr(base, '_get_attr', {}))

        slots = class_dict.get('__slots__', None)
        if slots is not None:
            # only add slots for new attr's; assume the layer above already slotted
            # if this layer is setting slots.
            class_dict['__slots__'] = tuple(
                sequences.iter_stable_unique(
                    itertools.chain(
                        slots,
                        set(new_functions).difference(existing)
                    )
                )
            )

        d = existing if class_dict.pop('__DynamicGetattrSetter_auto_inherit__', True) else {}
        d.update(new_functions)
        d.update(class_dict.pop('_get_attr', {}))
        class_dict['_get_attr'] = d
        class_dict.setdefault('__getattr__', dynamic_getattr_dict)

        return type.__new__(cls, name, bases, class_dict)



import argparse
import grp
import itertools
import locale
import os
import pwd
import re
import shlex
import shutil
import stat
from operator import itemgetter

from snakeoil.cli import arghparse
from snakeoil.compression import ArComp, ArCompError
from snakeoil.contexts import chdir
from snakeoil.decorators import coroutine
from snakeoil.iterables import partition
from snakeoil.osutils import pjoin
from snakeoil.process import spawn

from .. import os_data
from ..exceptions import PkgcoreException, PkgcoreUserException
from . import atom as atom_mod
from . import filter_env, portageq
from .misc import get_relative_dosym_target


class IpcError(PkgcoreException):
    """Generic IPC errors."""

    def __init__(self, msg='', code=1, name=None, **kwargs):
        super().__init__(msg, **kwargs)
        self.msg = msg
        self.code = code
        self.name = name
        self.ret = IpcCommand._encode_ret((code, msg))

    def __str__(self):
        if self.name:
            return f'{self.name}: {self.msg}'
        return self.msg


class IpcInternalError(IpcError):
    """IPC errors related to internal bugs."""


class IpcCommandError(IpcError, PkgcoreUserException):
    """IPC errors related to parsing arguments or running the command."""


class UnknownOptions(IpcCommandError):
    """Unknown options passed to IPC command."""

    def __init__(self, options):
        super().__init__(f"unknown options: {', '.join(map(repr, options))}")


class UnknownArguments(IpcCommandError):
    """Unknown arguments passed to IPC command."""

    def __init__(self, args):
        super().__init__(f"unknown arguments: {', '.join(map(repr, args))}")


class IpcArgumentParser(arghparse.ArgumentParser):
    """Raise IPC exception for argparse errors.

    Otherwise standard argparse prints the parser usage then outputs the error
    message to stderr.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, suppress=True, add_help=False, **kwargs)

    def error(self, msg):
        raise IpcCommandError(msg)


class IpcCommand:
    """Commands sent from the bash side of the ebuild daemon to run."""

    # argument parser for internal options
    parser = None
    # argument parser for command options/arguments
    arg_parser = None
    # override IPC name for error messages
    name = None

    def __init__(self, op):
        self.op = op
        self.pkg = op.pkg
        self.eapi = op.pkg.eapi
        self.observer = op.observer
        if self.name is None:
            self.name = self.__class__.__name__.lower()

    def __call__(self, ebd):
        self.opts = arghparse.Namespace()
        self.ebd = ebd
        ret = 0

        # read info from bash side
        nonfatal = self.read() == 'true'
        self.cwd = self.read()
        self.phase = self.read()
        options = shlex.split(self.read())
        args = self.read().strip('\0')
        args = args.split('\0') if args else []

        # parse args and run command
        with chdir(self.cwd):
            try:
                args = self.parse_args(options, args)
                ret = self.run(args)
            except IpcCommandError as e:
                if nonfatal:
                    ret = (e.code, e.msg)
                else:
                    raise IpcCommandError(msg=e.msg, code=e.code, name=self.name)
            except KeyboardInterrupt:
                raise
            except Exception as e:
                raise IpcInternalError('internal failure') from e

        # return completion status to the bash side
        self.write(self._encode_ret(ret))

    @staticmethod
    def _encode_ret(ret):
        """Encode exit status and any returned value to be sent back to the bash side."""
        if ret is None:
            return 0
        elif isinstance(ret, tuple):
            code, response = ret
            return f'{code}\x07{response}'
        elif isinstance(ret, (int, str)):
            return f'0\x07{ret}'
        raise TypeError(f'unsupported return status type: {type(ret)}')

    def parse_args(self, options, args):
        """Parse internal args passed from the bash side."""
        if self.parser is not None:
            _, unknown = self.parser.parse_known_args(options, namespace=self.opts)
            if unknown:
                raise UnknownOptions(unknown)

        if self.arg_parser is not None:
            # pull user options off the start of the argument list
            _, args = self.arg_parser.parse_known_optionals(args, namespace=self.opts)
            # parse remaining command arguments
            args, unknown = self.arg_parser.parse_known_args(args, namespace=self.opts)
            if unknown:
                raise UnknownArguments(unknown)
        return args

    def run(self, args):
        """Run the requested IPC command."""
        raise NotImplementedError

    def read(self):
        """Read a line from the ebuild daemon."""
        return self.ebd.read().strip()

    def write(self, data):
        """Write data to the ebuild daemon.

        Args:
            data: data to be sent to the bash side
        """
        self.ebd.write(data)

    def warn(self, msg):
        """Output warning message.

        Args:
            msg: message to be output
        """
        self.observer.warn(f'{self.name}: {msg}')
        self.observer.flush()


def _parse_group(group):
    try:
        return grp.getgrnam(group).gr_gid
    except KeyError:
        pass
    return int(group)


def _parse_user(user):
    try:
        return pwd.getpwnam(user).pw_uid
    except KeyError:
        pass
    return int(user)


def _parse_mode(mode):
    try:
        return int(mode, 8)
    except ValueError:
        return None


def command_options(s):
    """Split string of command line options into list."""
    return shlex.split(s)


def existing_path(path):
    """Check if a given path exists (allows broken symlinks)."""
    if not os.path.lexists(path):
        raise argparse.ArgumentTypeError(f'nonexistent path: {path!r}')
    return path


class _InstallWrapper(IpcCommand):
    """Python wrapper for commands using `install`."""

    parser = IpcArgumentParser()
    parser.add_argument('--dest', default='/')
    parser.add_argument('--insoptions', type=command_options)
    parser.add_argument('--diroptions', type=command_options)

    # defaults options for file and dir install actions
    insoptions_default = ''
    diroptions_default = ''

    # supported install command options
    install_parser = IpcArgumentParser()
    install_parser.add_argument('-g', '--group', default=-1, type=_parse_group)
    install_parser.add_argument('-o', '--owner', default=-1, type=_parse_user)
    install_parser.add_argument('-m', '--mode', default=0o755, type=_parse_mode)
    install_parser.add_argument('-p', '--preserve-timestamps', action='store_true')

    arg_parser = IpcArgumentParser()
    arg_parser.add_argument('targets', nargs='+', type=existing_path)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parser.set_defaults(
            insoptions=self.insoptions_default,
            diroptions=self.diroptions_default)

        # initialize file/dir creation coroutines
        self.install = self._install().send
        self.install_dirs = self._install_dirs().send
        self.install_symlinks = self._install_symlinks().send
        self.install_from_dirs = self._install_from_dirs().send

    def parse_args(self, *args, **kwargs):
        args = super().parse_args(*args, **kwargs)
        self.parse_install_options()
        return args

    def parse_install_options(self):
        """Parse install command options."""
        self.insoptions = arghparse.Namespace()
        self.diroptions = arghparse.Namespace()
        if self.opts.insoptions:
            if not self._parse_install_options(self.opts.insoptions, self.insoptions):
                self.install = self._install_cmd().send
        if self.opts.diroptions:
            if not self._parse_install_options(self.opts.diroptions, self.diroptions):
                self.install_dirs = self._install_dirs_cmd().send

    def _parse_install_options(self, options, namespace):
        """Internal install command option parser.

        Args:
            options: list of options to parse
            namespace: argparse namespace to populate
        Returns:
            True when all options are handled,
            otherwise False if unknown/unhandled options exist.
        """
        opts, unknown = self.install_parser.parse_known_args(options, namespace=namespace)
        if unknown or opts.mode is None:
            msg = "falling back to 'install'"
            if unknown:
                msg += f": unhandled options: {' '.join(map(repr, unknown))}"
            self.warn(msg)
            return False
        return True

    def run(self, args):
        try:
            dest_dir = pjoin(self.op.ED, self.opts.dest.lstrip(os.path.sep))
            os.makedirs(dest_dir, exist_ok=True)
        except OSError as e:
            raise IpcCommandError(
                f'failed creating dir: {dest_dir!r}: {e.strerror}')
        self._install_targets(args.targets)

    def _prefix_targets(self, targets, files=True):
        """Prepend targets being installed with the destination path."""
        dest_dir = self.opts.dest.lstrip(os.path.sep)
        if files:
            return (
                (s, pjoin(self.op.ED, dest_dir, d.lstrip(os.path.sep)))
                for s, d in targets)
        return (
            pjoin(self.op.ED, dest_dir, d.lstrip(os.path.sep))
            for d in targets)

    def _install_targets(self, targets):
        """Install targets.

        Args:
            targets: files/symlinks/dirs/etc to install
        """
        self.install((f, os.path.basename(f)) for f in targets)

    @coroutine
    def _install_from_dirs(self):
        """Install all targets under given directories.

        Args:
            iterable of directories to install from
        """
        while True:
            dirs = (yield)
            for d in dirs:
                base_dir = os.path.basename(d.rstrip(os.path.sep))
                for dirpath, dirnames, filenames in os.walk(d):
                    dest_dir = os.path.normpath(pjoin(base_dir, os.path.relpath(dirpath, d)))
                    self.install_dirs([dest_dir])
                    for dirname in dirnames:
                        source = pjoin(dirpath, dirname)
                        if os.path.islink(source):
                            dest = pjoin(dest_dir, dirname)
                            self.install_symlinks([(source, dest)])
                    if filenames:
                        self.install(
                            (pjoin(dirpath, f), pjoin(dest_dir, f))
                            for f in filenames
                        )

    @staticmethod
    def _set_attributes(opts, path):
        """Set file attributes on a given path.

        Args:
            path: file/directory path
        """
        try:
            if opts.owner != -1 or opts.group != -1:
                os.lchown(path, opts.owner, opts.group)
            if opts.mode is not None and not os.path.islink(path):
                os.chmod(path, opts.mode)
        except OSError as e:
            raise IpcCommandError(
                f'failed setting file attributes: {path!r}: {e.strerror}')

    @staticmethod
    def _set_timestamps(source_stat, dest):
        """Apply timestamps from source_stat to dest.

        Args:
            source_stat: stat result for the source file
            dest: path to the dest file
        """
        os.utime(dest, ns=(source_stat.st_atime_ns, source_stat.st_mtime_ns))

    def _is_install_allowed(self, source, source_stat, dest):
        """Determine if installing source into dest should work.

        This aims to aid compatibility with the `install` command.

        Args:
            source: path to the source file
            source_stat: stat result for the source file, using stat()
                rather than lstat(), in order to match the `install`
                command
            dest: path to the dest file
        Raises:
            IpcCommandError on failure
        Returns:
            True if the install should succeed
        """
        # matching `install` command, use stat() for source and lstat() for dest
        try:
            dest_lstat = os.lstat(dest)
        except FileNotFoundError:
            # installing file to a new path
            return True
        except OSError as e:
            raise IpcCommandError(f'cannot stat {dest!r}: {e.strerror}')

        # installing symlink
        if stat.S_ISLNK(dest_lstat.st_mode):
            return True

        # source file and dest file are different
        if not os.path.samestat(source_stat, dest_lstat):
            return True

        # installing hardlink if source and dest are different
        if (dest_lstat.st_nlink > 1 and os.path.realpath(source) != os.path.realpath(dest)):
            return True

        raise IpcCommandError(f'{source!r} and {dest!r} are identical')

    @coroutine
    def _install(self):
        """Install files.

        Args:
            iterable of (source, dest) tuples of files to install
        Raises:
            IpcCommandError on failure
        """
        while True:
            files = (yield)
            # TODO: skip/warn installing empty files
            for source, dest in self._prefix_targets(files):
                try:
                    sstat = os.stat(source)
                except OSError as e:
                    raise IpcCommandError(f'cannot stat {source!r}: {e.strerror}')

                self._is_install_allowed(source, sstat, dest)

                # matching `install` command, remove dest before file install
                try:
                    os.unlink(dest)
                except FileNotFoundError:
                    pass
                except OSError as e:
                    raise IpcCommandError(f'failed removing file: {dest!r}: {e.strerror}')

                try:
                    shutil.copyfile(source, dest, follow_symlinks=False)
                    if self.insoptions:
                        self._set_attributes(self.insoptions, dest)
                        if self.insoptions.preserve_timestamps:
                            self._set_timestamps(sstat, dest)
                except OSError as e:
                    raise IpcCommandError(
                        f'failed copying file: {source!r} to {dest!r}: {e.strerror}')

    @coroutine
    def _install_cmd(self):
        """Install files using `install` command.

        Args:
            iterable of (source, dest) tuples of files to install
        Raises:
            IpcCommandError on failure
        """
        while True:
            files = (yield)

            # `install` forcibly resolves symlinks so split them out
            files, symlinks = partition(files, predicate=lambda x: os.path.islink(x[0]))
            self.install_symlinks(symlinks)

            # group and install sets of files by destination to decrease `install` calls
            files = sorted(self._prefix_targets(files), key=itemgetter(1))
            for dest, files_group in itertools.groupby(files, itemgetter(1)):
                sources = list(path for path, _ in files_group)
                command = ['install'] + self.opts.insoptions + sources + [dest]
                ret, output = spawn.spawn_get_output(command, collect_fds=(2,))
                if not ret:
                    raise IpcCommandError('\n'.join(output), code=ret)

    @coroutine
    def _install_dirs(self):
        """Create directories.

        Args:
            iterable of paths where directories should be created
        Raises:
            IpcCommandError on failure
        """
        while True:
            dirs = (yield)
            try:
                for d in self._prefix_targets(dirs, files=False):
                    os.makedirs(d, exist_ok=True)
                    if self.diroptions:
                        self._set_attributes(self.diroptions, d)
            except OSError as e:
                raise IpcCommandError(f'failed creating dir: {d!r}: {e.strerror}')

    @coroutine
    def _install_dirs_cmd(self):
        """Create directories using `install` command.

        Args:
            iterable of paths where directories should be created
        Raises:
            IpcCommandError on failure
        """
        while True:
            dirs = (yield)
            dirs = self._prefix_targets(dirs, files=False)
            command = ['install', '-d'] + self.opts.diroptions + list(dirs)
            ret, output = spawn.spawn_get_output(command, collect_fds=(2,))
            if not ret:
                raise IpcCommandError('\n'.join(output), code=ret)

    @coroutine
    def _install_symlinks(self):
        """Install iterable of symlinks.

        Args:
            iterable of (path, target dir) tuples of symlinks to install
        Raises:
            IpcCommandError on failure
        """
        while True:
            symlinks = (yield)
            try:
                for symlink, dest in self._prefix_targets(symlinks):
                    os.symlink(os.readlink(symlink), dest)
            except OSError as e:
                raise IpcCommandError(
                    f'failed creating symlink: {symlink!r} -> {dest!r}: {e.strerror}')


class Doins(_InstallWrapper):
    """Python wrapper for doins."""

    arg_parser = IpcArgumentParser(parents=(_InstallWrapper.arg_parser,))
    arg_parser.add_argument('-r', dest='recursive', action='store_true')

    def _install_targets(self, targets):
        files, dirs = partition(targets, predicate=os.path.isdir)
        if self.opts.recursive:
            self.install_from_dirs(dirs)
        self.install((f, os.path.basename(f)) for f in files)


class Dodoc(_InstallWrapper):
    """Python wrapper for dodoc."""

    insoptions_default = '-m0644'

    arg_parser = IpcArgumentParser(parents=(_InstallWrapper.arg_parser,))
    arg_parser.add_argument('-r', dest='recursive', action='store_true')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.allow_recursive = self.eapi.options.dodoc_allow_recursive

    def _install_targets(self, targets):
        files, dirs = partition(targets, predicate=os.path.isdir)
        # TODO: add peekable class for iterables to avoid list conversion
        dirs = list(dirs)
        if dirs:
            if self.opts.recursive and self.allow_recursive:
                self.install_from_dirs(dirs)
            else:
                missing_option = ', missing -r option?' if self.allow_recursive else ''
                raise IpcCommandError(f'{dirs[0]!r} is a directory{missing_option}')
        self.install((f, os.path.basename(f)) for f in files)


class Doinfo(_InstallWrapper):
    """Python wrapper for doinfo."""

    insoptions_default = '-m0644'


class Dodir(_InstallWrapper):
    """Python wrapper for dodir."""

    diroptions_default = '-m0755'

    arg_parser = IpcArgumentParser()
    arg_parser.add_argument('targets', nargs='+')

    def run(self, args):
        self.install_dirs(args.targets)


class Keepdir(Dodir):
    """Python wrapper for keepdir."""

    def run(self, args):
        # create dirs
        super().run(args)

        # create stub files
        filename = f'.keep_{self.pkg.category}_{self.pkg.PN}-{self.pkg.slot}'
        for x in args.targets:
            path = pjoin(self.op.ED, x.lstrip(os.path.sep), filename)
            open(path, 'w').close()


class Doexe(_InstallWrapper):
    """Python wrapper for doexe."""


class Dobin(_InstallWrapper):
    """Python wrapper for dobin."""

    def parse_install_options(self, *args, **kwargs):
        # TODO: fix this to be prefix aware at some point
        self.opts.insoptions = [
            '-m0755', f'-g{os_data.root_gid}', f'-o{os_data.root_uid}']
        return super().parse_install_options(*args, **kwargs)


class Dosbin(Dobin):
    """Python wrapper for dosbin."""


class Dolib(_InstallWrapper):
    """Python wrapper for dolib."""


class Dolib_so(Dolib):
    """Python wrapper for dolib.so."""

    name = 'dolib.so'


class Dolib_a(Dolib):
    """Python wrapper for dolib.a."""

    name = 'dolib.a'


class _Symlink(_InstallWrapper):

    arg_parser = IpcArgumentParser()
    arg_parser.add_argument('source')
    arg_parser.add_argument('target')

    def run(self, args):
        dest_dir = args.target.rsplit(os.path.sep, 1)[0]
        if dest_dir != args.target:
            self.install_dirs([dest_dir])

        target = pjoin(self.op.ED, args.target.lstrip(os.path.sep))
        with chdir(self.op.ED):
            try:
                try:
                    self._link(args.source, target)
                except FileExistsError:
                    # overwrite target if it exists
                    os.unlink(target)
                    self._link(args.source, target)
            except OSError as e:
                raise IpcCommandError(
                    f'failed creating link: {args.source!r} -> {args.target!r}: {e.strerror}')


class Dosym(_Symlink):
    """Python wrapper for dosym."""

    _link = os.symlink
    arg_parser = IpcArgumentParser(parents=(_Symlink.arg_parser,))
    arg_parser.add_argument('-r', dest='relative', action='store_true')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dosym_relative = self.eapi.options.dosym_relative

    def run(self, args):
        target = args.target
        if (target.endswith(os.path.sep) or
                (os.path.isdir(target) and not os.path.islink(target))):
            # bug 379899
            raise IpcCommandError(f'missing filename target: {target!r}')

        if self.opts.relative:
            if not self.dosym_relative:
                raise IpcCommandError(f'-r not permitted in EAPI {self.eapi}')
            if not os.path.isabs(args.source):
                raise IpcCommandError('-r is only meaningful with absolute paths')
            args.source = get_relative_dosym_target(args.source, target)

        super().run(args)


class Dohard(_Symlink):
    """Python wrapper for dosym."""

    _link = os.link


class Doman(_InstallWrapper):
    """Python wrapper for doman."""

    insoptions_default = '-m0644'

    arg_parser = IpcArgumentParser(parents=(_InstallWrapper.arg_parser,))
    arg_parser.add_argument('-i18n', action='store_true', default='')

    detect_lang_re = re.compile(r'^(\w+)\.([a-z]{2}([A-Z]{2})?)\.(\w+)$')
    valid_mandir_re = re.compile(r'man[0-9n](f|p|pm)?$')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.language_detect = self.eapi.options.doman_language_detect
        self.language_override = self.eapi.options.doman_language_override

    def _install_targets(self, targets):
        dirs = set()
        for x in targets:
            basename = os.path.basename(x)
            ext = os.path.splitext(basename)[1]

            if self.eapi.archive_exts_regex.match(ext):
                # TODO: uncompress/warn?
                ext = os.path.splitext(basename.rsplit('.', 1)[0])[1]

            name = basename
            mandir = f'man{ext[1:]}'

            if self.language_override and self.opts.i18n:
                mandir = pjoin(self.opts.i18n, mandir)
            elif self.language_detect:
                match = self.detect_lang_re.match(basename)
                if match:
                    name = f'{match.group(1)}.{match.group(4)}'
                    mandir = pjoin(match.group(2), mandir)

            if self.valid_mandir_re.match(os.path.basename(mandir)):
                if mandir not in dirs:
                    self.install_dirs([mandir])
                    dirs.add(mandir)
                self.install([(x, pjoin(mandir, name))])
            else:
                raise IpcCommandError(f'invalid man page: {x}')


class Domo(_InstallWrapper):
    """Python wrapper for domo."""

    insoptions_default = '-m0644'

    def _install_targets(self, targets):
        dirs = set()
        for x in targets:
            d = pjoin(os.path.splitext(os.path.basename(x))[0], 'LC_MESSAGES')
            if d not in dirs:
                self.install_dirs([d])
                dirs.add(d)
            self.install([(x, pjoin(d, f'{self.pkg.PN}.mo'))])


class Dohtml(_InstallWrapper):
    """Python wrapper for dohtml."""

    insoptions_default = '-m0644'

    arg_parser = IpcArgumentParser(parents=(_InstallWrapper.arg_parser,))
    arg_parser.add_argument('-r', dest='recursive', action='store_true')
    arg_parser.add_argument('-V', dest='verbose', action='store_true')
    arg_parser.add_argument('-A', dest='extra_allowed_file_exts', action='csv', default=[])
    arg_parser.add_argument('-a', dest='allowed_file_exts', action='csv', default=[])
    arg_parser.add_argument('-f', dest='allowed_files', action='csv', default=[])
    arg_parser.add_argument('-x', dest='excluded_dirs', action='csv', default=[])
    arg_parser.add_argument('-p', dest='doc_prefix', default='')

    # default allowed file extensions
    default_allowed_file_exts = ('css', 'gif', 'htm', 'html', 'jpeg', 'jpg', 'js', 'png')

    def parse_args(self, *args, **kwargs):
        args = super().parse_args(*args, **kwargs)
        self.opts.dest = pjoin(self.opts.dest, self.opts.doc_prefix.lstrip(os.path.sep))

        if not self.opts.allowed_file_exts:
            self.opts.allowed_file_exts = list(self.default_allowed_file_exts)
        self.opts.allowed_file_exts.extend(self.opts.extra_allowed_file_exts)

        self.opts.allowed_file_exts = set(self.opts.allowed_file_exts)
        self.opts.excluded_dirs = set(self.opts.excluded_dirs)
        self.opts.allowed_files = set(self.opts.allowed_files)

        if self.opts.verbose:
            self.observer.write(str(self), autoline=True)
            self.observer.flush()

        return args

    def __str__(self):
        msg = ['dohtml:', f'  Installing to: {self.opts.dest}']
        if self.opts.allowed_file_exts:
            msg.append(
                f"  Allowed extensions: {', '.join(sorted(self.opts.allowed_file_exts))}")
        if self.opts.excluded_dirs:
            msg.append(
                f"  Allowed extensions: {', '.join(sorted(self.opts.allowed_file_exts))}")
        if self.opts.allowed_files:
            msg.append(
                f"  Allowed files: {', '.join(sorted(self.opts.allowed_files))}")
        if self.opts.doc_prefix:
            msg.append(f"  Document prefix: {self.opts.doc_prefix!r}")
        return '\n'.join(msg)

    def _allowed_file(self, path):
        """Determine if a file is allowed to be installed."""
        basename = os.path.basename(path)
        ext = os.path.splitext(basename)[1][1:]
        return (ext in self.opts.allowed_file_exts or basename in self.opts.allowed_files)

    def _install_targets(self, targets):
        files, dirs = partition(targets, predicate=os.path.isdir)
        # TODO: add peekable class for iterables to avoid list conversion
        dirs = list(dirs)
        if dirs:
            if self.opts.recursive:
                dirs = (d for d in dirs if d not in self.opts.excluded_dirs)
                self.install_from_dirs(dirs)
            else:
                raise IpcCommandError(f'{dirs[0]!r} is a directory, missing -r option?')
        self.install((f, os.path.basename(f)) for f in files if self._allowed_file(f))


class _AlterFiles(IpcCommand):

    arg_parser = IpcArgumentParser()
    arg_parser.add_argument('-x', dest='excludes', action='store_true')
    arg_parser.add_argument('targets', nargs='+')

    default_includes = ()
    default_excludes = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.includes = set(self.default_includes)
        self.excludes = set(self.default_excludes)

    def run(self, args):
        if self.opts.excludes:
            self.excludes.update(args.targets)
        else:
            self.includes.update(args.targets)


class Docompress(_AlterFiles):
    """Python wrapper for docompress."""

    default_includes = ('/usr/share/doc', '/usr/share/info', '/usr/share/man')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.excludes = {f'/usr/share/doc/{self.pkg.PF}/html'}


class Dostrip(_AlterFiles):
    """Python wrapper for dostrip."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'strip' not in self.pkg.restrict:
            self.includes = {'/'}


class _QueryCmd(IpcCommand):

    arg_parser = IpcArgumentParser()
    arg_parser.add_argument('atom', type=atom_mod.atom)

    # >= EAPI 5
    host_root_parser = IpcArgumentParser()
    host_root_parser.add_argument('--host-root', action='store_true')

    # >= EAPI 7
    query_deps_parser = IpcArgumentParser()
    dep_opts = query_deps_parser.add_mutually_exclusive_group()
    dep_opts.add_argument('-b', dest='bdepend', action='store_true')
    dep_opts.add_argument('-d', dest='depend', action='store_true')
    dep_opts.add_argument('-r', dest='rdepend', action='store_true')

    def parse_args(self, options, args):
        # parse EAPI specific optionals then remaining args
        if self.eapi.options.query_host_root:
            _, args = self.host_root_parser.parse_known_optionals(args, namespace=self.opts)
        elif self.eapi.options.query_deps:
            _, args = self.query_deps_parser.parse_known_optionals(args, namespace=self.opts)
        args = super().parse_args(options, args)

        root = None
        self.opts.domain = self.op.domain

        if self.eapi.options.query_host_root and self.opts.host_root:
            root = '/'
        elif self.eapi.options.query_deps:
            if self.opts.bdepend:
                if self.pkg.eapi.options.prefix_capable:
                    # not using BROOT as that's only defined in src_* phases
                    root = pjoin('/', self.op.env['EPREFIX'])
                else:
                    root = '/'
            elif self.opts.depend:
                if self.pkg.eapi.options.prefix_capable:
                    root = self.op.env['ESYSROOT']
                else:
                    root = self.op.env['SYSROOT']
            else:
                if self.pkg.eapi.options.prefix_capable:
                    root = self.op.env['EROOT']
                else:
                    root = self.op.env['ROOT']

        # TODO: find domain from given path, pointless until full prefix support works
        if root and root != self.opts.domain.root:
            raise IpcCommandError('prefix support not implemented yet')

        return args


class Has_Version(_QueryCmd):
    """Python wrapper for has_version."""

    def run(self, args):
        if args.atom in self.opts.domain.all_installed_repos:
            return 0
        return 1


class Best_Version(_QueryCmd):
    """Python wrapper for best_version."""

    def run(self, args):
        return portageq._best_version(self.opts.domain, args.atom)


class Eapply(IpcCommand):
    """Python wrapper for eapply."""

    arg_parser = IpcArgumentParser()
    arg_parser.add_argument('targets', nargs='+', type=existing_path)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.patch_cmd = ['patch', '-p1', '-f', '-s', '-g0', '--no-backup-if-mismatch']
        self.patch_opts = []

    def _parse_patch_opts(self, args):
        patch_opts = []
        files = []
        for i, arg in enumerate(args):
            if arg == '--':
                if files:
                    raise IpcCommandError('options must be specified before file arguments')
                files = args[i + 1:]
                break
            elif arg.startswith('-'):
                if files:
                    raise IpcCommandError('options must be specified before file arguments')
                patch_opts.append(arg)
            else:
                files.append(arg)
        return files, patch_opts

    def _find_patches(self, args):
        for path in args:
            if os.path.isdir(path):
                for root, _dirs, files in os.walk(path):
                    patches = [
                        pjoin(root, f) for f in sorted(files, key=locale.strxfrm)
                        if f.endswith(('.diff', '.patch'))]
                    if not patches:
                        raise IpcCommandError(f'no patches in directory: {path!r}')
                    yield path, patches
            else:
                yield None, [path]

    def parse_args(self, options, args):
        args, self.patch_opts = self._parse_patch_opts(args)
        args = super().parse_args(options, ['--'] + args)
        return self._find_patches(args.targets)

    def run(self, args, user=False):
        if user:
            patch_type = 'user patches'
            output_func = self.observer.warn
        else:
            patch_type = 'patches'
            output_func = self.observer.info

        spawn_kwargs = {'collect_fds': (1, 2)}
        if self.op.userpriv:
            spawn_kwargs['uid'] = os_data.portage_uid
            spawn_kwargs['gid'] = os_data.portage_gid

        for path, patches in args:
            prefix = ''
            if path is not None:
                output_func(f'Applying {patch_type} from {path!r}:')
                prefix = '  '
            for patch in patches:
                if path is None:
                    output_func(f'{prefix}Applying {os.path.basename(patch)}...')
                else:
                    output_func(f'{prefix}{os.path.basename(patch)}...')
                self.observer.flush()
                try:
                    with open(patch) as f:
                        ret, output = spawn.spawn_get_output(
                            self.patch_cmd + self.patch_opts,
                            fd_pipes={0: f.fileno()}, **spawn_kwargs)
                    if ret:
                        filename = os.path.basename(patch)
                        msg = f'applying {filename!r} failed: {output[0]}'
                        raise IpcCommandError(msg, code=ret)
                except OSError as e:
                    raise IpcCommandError(
                        f'failed reading patch file: {patch!r}: {e.strerror}')


class Eapply_User(IpcCommand):
    """Python wrapper for eapply_user."""

    # stub parser so any arguments are flagged as errors
    arg_parser = IpcArgumentParser()

    def run(self, args):
        if self.pkg.user_patches:
            self.op._ipc_helpers['eapply'].run(self.pkg.user_patches, user=True)

        # create marker to skip additionals calls
        patches = itertools.chain.from_iterable(
            files for _, files in self.pkg.user_patches)
        with open(pjoin(self.op.env['T'], '.user_patches_applied'), 'w') as f:
            f.write('\n'.join(patches))


class Unpack(IpcCommand):

    arg_parser = IpcArgumentParser()
    arg_parser.add_argument('targets', nargs='+')

    _file_mode = (
        stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH
        | stat.S_IWUSR
        & ~stat.S_IWGRP & ~stat.S_IWOTH
    )
    _dir_mode = _file_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH

    def parse_args(self, *args, **kwargs):
        args = super().parse_args(*args, **kwargs)
        self.opts.distdir = self.op.env['DISTDIR']
        args.targets = self._filter_targets(args.targets)
        return args

    def _filter_targets(self, targets):
        for archive in targets:
            if os.path.sep not in archive:
                # regular filename get prefixed with ${DISTDIR}
                srcdir = self.opts.distdir
            elif archive.startswith('./'):
                # relative paths get passed through
                srcdir = ''
            else:
                srcdir = self.opts.distdir

                # >= EAPI 6 allows absolute paths
                if self.eapi.options.unpack_absolute_paths:
                    srcdir = ''
                    if archive.startswith(self.opts.distdir):
                        self.warn(
                            f'argument contains redundant ${{DISTDIR}}: {archive!r}')
                elif archive.startswith(self.opts.distdir):
                    raise IpcCommandError(
                        f'arguments must not begin with ${{DISTDIR}}: {archive!r}')
                elif archive.startswith(os.path.sep):
                    raise IpcCommandError(
                        f'arguments must not be absolute paths: {archive!r}')
                else:
                    raise IpcCommandError(
                        'relative paths must be prefixed with '
                        f"'./' in EAPI {self.eapi}")

            path = pjoin(srcdir, archive)
            if not os.path.exists(path):
                raise IpcCommandError(f'nonexistent file: {archive!r}')
            elif os.stat(path).st_size == 0:
                raise IpcCommandError(f'empty file: {archive!r}')

            match = self.eapi.archive_exts_regex.search(archive)
            if not match:
                self.warn(f'skipping unrecognized file format: {archive!r}')
                continue
            ext = match.group(1)

            yield archive, ext, path

    def run(self, args):
        spawn_kwargs = {}
        if self.op.userpriv and self.phase == 'unpack':
            spawn_kwargs['uid'] = os_data.portage_uid
            spawn_kwargs['gid'] = os_data.portage_gid

        for filename, ext, source in args.targets:
            self.observer.write(f'>>> Unpacking {filename} to {self.cwd}', autoline=True)
            self.observer.flush()
            dest = pjoin(self.cwd, os.path.basename(filename[:-len(ext)]))
            try:
                target = ArComp(source, ext=ext)
                target.unpack(dest=dest, **spawn_kwargs)
            except ArCompError as e:
                raise IpcCommandError(str(e), code=e.code)

        for dirpath, dirnames, filenames in os.walk(self.cwd):
            dirs = ((self._dir_mode, x) for x in dirnames)
            files = ((self._file_mode, x) for x in filenames)
            for mode, f in itertools.chain.from_iterable((dirs, files)):
                path = pjoin(dirpath, f)
                current_mode = os.lstat(path).st_mode
                if not stat.S_ISLNK(current_mode):
                    os.chmod(path, current_mode | mode)


class FilterEnv(IpcCommand):

    arg_parser = IpcArgumentParser()
    filtering = arg_parser.add_argument_group("Environment filtering options")
    filtering.add_argument(
        '-V', '--var-match', action='store_true', default=False,
        help="invert the filtering- instead of removing a var if it matches "
        "remove all vars that do not match")
    filtering.add_argument(
        '-F', '--func-match', action='store_true', default=False,
        help="invert the filtering- instead of removing a function if it matches "
        "remove all functions that do not match")
    filtering.add_argument(
        '-f', '--funcs', action='csv',
        help="comma separated list of regexes to match function names against for filtering")
    filtering.add_argument(
        '-v', '--vars', action='csv',
        help="comma separated list of regexes to match variable names against for filtering")
    arg_parser.add_argument('files', nargs=2)

    def run(self, args):
        src_path, dest_path = args.files
        with open(src_path) as src, open(dest_path, 'wb') as dest:
            filter_env.main_run(
                dest, src.read(), args.vars, args.funcs,
                args.var_match, args.func_match)



"""
template for fs based backends
"""

__all__ = ("FsBased",)

import os

from snakeoil.osutils import ensure_dirs, pjoin

from ..os_data import portage_gid
from . import base


class FsBased(base):
    """Template wrapping fs needed options.

    Provides _ensure_access as a way to attempt to ensure files have
    the specified owners/perms.
    """

    def __init__(self, location, label=None, **config):
        """
        throws InitializationError if needs args aren't specified

        :keyword gid: defaults to :obj:`pkgcore.os_data.portage_gid`,
            gid to force all entries to
        :keyword perms: defaults to 0665, mode to force all entries to"""

        for x, y in (("gid", portage_gid), ("perms", 0o664)):
            if x in config:
                setattr(self, f'_{x}', config[x])
                del config[x]
            else:
                setattr(self, f'_{x}', y)
        super().__init__(**config)

        if label is not None:
            location = pjoin(location, label.lstrip(os.path.sep))

        self.location = location

        self._mtime_used = 'mtime' == self.chf_type

    __init__.__doc__ = "\n".join(
        x.lstrip() for x in __init__.__doc__.split("\n") + [
            y.lstrip().replace("@param", "@keyword")
            for y in base.__init__.__doc__.split("\n")
            if "@param" in y])

    def _ensure_access(self, path, mtime=None):
        """Ensure access to a path.

        :param mtime: if specified change mtime to this value.
        :return: C{False} if unable to guarantee access, C{True} otherwise.
        """
        try:
            os.chown(path, -1, self._gid)
            os.chmod(path, self._perms)
            if mtime is not None:
                mtime = int(mtime)
                os.utime(path, (mtime, mtime))
        except EnvironmentError:
            return False
        return True

    def _ensure_dirs(self, path=None):
        """Make sure a path relative to C{self.location} exists."""
        if path is not None:
            path = pjoin(self.location, os.path.dirname(path))
        else:
            path = self.location
        return ensure_dirs(path, mode=0o775, minimal=False)



"""
ini based configuration format
"""

__all__ = ("config_from_file",)

import configparser

from snakeoil import mappings

from . import basics, errors


class CaseSensitiveConfigParser(configparser.ConfigParser):
    def optionxform(self, val):
        return val


def config_from_file(file_obj):
    """
    generate a config dict

    :param file_obj: file protocol instance
    :return: :obj:`snakeoil.mappings.LazyValDict` instance
    """
    cparser = CaseSensitiveConfigParser()
    try:
        cparser.read_file(file_obj)
    except configparser.ParsingError as e:
        raise errors.ParsingError(f'while parsing {file_obj}', e) from e
    def get_section(section):
        return basics.ConfigSectionFromStringDict(dict(cparser.items(section)))
    return mappings.LazyValDict(cparser.sections, get_section)



from pkgcore.resolver.choice_point import choice_point
from pkgcore.restrictions.boolean import AndRestriction, OrRestriction
from snakeoil.test import TestCase


class fake_package:

    def __init__(self, **kwds):
        for k, v in (
                ("bdepend", AndRestriction()),
                ("depend", AndRestriction()),
                ("rdepend", AndRestriction()),
                ("pdepend", AndRestriction()),
                ("idepend", AndRestriction()),
                ("slot", 0),
                ("key", None),
                ("marker", None)):
            setattr(self, k, kwds.get(k, v))

class TestChoicePoint(TestCase):

    @staticmethod
    def gen_choice_point():
        return choice_point("asdf", [
            fake_package(marker=1, depend=OrRestriction(
                        "ordep1", "ordep2", "dependordep"),
                rdepend=AndRestriction(
                        OrRestriction("ordep1", "andordep2"),
                        "anddep1", "anddep2", "pkg1and"),
                pdepend=OrRestriction("prdep1", "or3")),
            fake_package(marker=2, depend=AndRestriction(
                        "anddep1", "anddep2"),
                rdepend=OrRestriction("or1", "or2"),
                pdepend=OrRestriction("prdep1", "or3"))])

    def test_depend_rdepend_stepping(self):
        c = self.gen_choice_point()
        self.assertEqual(c.depend, [["ordep1", "ordep2", "dependordep"]])
        self.assertEqual(
            sorted(c.rdepend),
            sorted([['anddep1'], ['anddep2'], ['ordep1', 'andordep2'],
                    ['pkg1and']]))
        c.reduce_atoms("ordep1")
        self.assertEqual(c.depend, [['ordep2', 'dependordep']])
        self.assertEqual(
            sorted(c.rdepend),
            sorted([['anddep1'], ['anddep2'], ['andordep2'], ['pkg1and']]))
        c.reduce_atoms("pkg1and")
        c.reduce_atoms("or1")
        self.assertEqual(c.rdepend, [["or2"]])
        c.reduce_atoms("prdep1")
        self.assertEqual(c.depend, [['anddep1'], ['anddep2']])
        self.assertEqual(c.pdepend, [["or3"]])
        c.reduce_atoms("or3")
        self.assertRaises(IndexError, lambda: c.depend)

    def test_current_pkg(self):
        c = self.gen_choice_point()
        self.assertEqual(c.current_pkg.marker, 1)
        c.reduce_atoms("pkg1and")
        self.assertEqual(c.current_pkg.marker, 2)

    def test_reduce(self):
        c = self.gen_choice_point()
        self.assertEqual(c.current_pkg.marker, 1)
        self.assertEqual(c.reduce_atoms("dependordep"), False)
        self.assertEqual(c.reduce_atoms("ordep2"), False)
        self.assertEqual(c.reduce_atoms("ordep1"), True)
        self.assertEqual(c.current_pkg.marker, 2)
        c = self.gen_choice_point()
        self.assertEqual(c.reduce_atoms("anddep2"), True)
        c = self.gen_choice_point()
        c.reduce_atoms("anddep1")
        self.assertRaises(IndexError, lambda: c.depend)
        self.assertRaises(IndexError, lambda: c.rdepend)
        self.assertRaises(IndexError, lambda: c.pdepend)

    def test_nonzero(self):
        c = self.gen_choice_point()
        self.assertEqual(bool(c), True)
        self.assertEqual(c.current_pkg.marker, 1)
        c.reduce_atoms("anddep1")
        self.assertEqual(bool(c), False)



__all__ = ("tree", "ConfiguredTree")

import errno
import os
import stat
from functools import partial

from snakeoil import data_source
from snakeoil.fileutils import readfile
from snakeoil.mappings import IndeterminantDict
from snakeoil.osutils import listdir_dirs, pjoin

from ..config.hint import ConfigHint
from ..ebuild import ebd, ebuild_built
from ..ebuild.cpv import VersionedCPV
from ..ebuild.errors import InvalidCPV
from ..log import logger
from ..package import base as pkg_base
from ..repository import errors, prototype, wrapper
from . import repo_ops
from .contents import ContentsFile


class tree(prototype.tree):
    """Repository for packages installed on the filesystem."""

    livefs = True
    configured = False
    configurables = ("domain", "settings")
    package_factory = staticmethod(ebuild_built.generate_new_factory)
    operations_kls = repo_ops.operations

    pkgcore_config_type = ConfigHint(
        {'location': 'str',
         'cache_location': 'str', 'repo_id': 'str',
         'disable_cache': 'bool'},
        typename='repo')

    def __init__(self, location, cache_location=None, repo_id='vdb',
                 disable_cache=False):
        super().__init__(frozen=False)
        self.repo_id = repo_id
        self.location = location
        if disable_cache:
            cache_location = None
        elif cache_location is None:
            cache_location = pjoin("/var/cache/edb/dep", location.lstrip("/"))
        self.cache_location = cache_location
        self._versions_tmp_cache = {}
        try:
            st = os.stat(self.location)
            if not stat.S_ISDIR(st.st_mode):
                raise errors.InitializationError(
                    f"base not a dir: {self.location!r}")
            elif not st.st_mode & (os.X_OK | os.R_OK):
                raise errors.InitializationError(
                    f"base lacks read/executable: {self.location!r}")
        except FileNotFoundError:
            pass
        except OSError as e:
            raise errors.InitializationError(f'lstat failed on base: {self.location!r}') from e

        self.package_class = self.package_factory(self)

    def configure(self, *args):
        return ConfiguredTree(self, *args)

    def _get_categories(self, *optional_category):
        # return if optional_category is passed... cause it's not yet supported
        if optional_category:
            return {}
        try:
            try:
                return tuple(x for x in listdir_dirs(self.location) if not
                             x.startswith('.'))
            except EnvironmentError as e:
                raise KeyError(f"failed fetching categories: {e}") from e
        finally:
            pass

    def _get_packages(self, category):
        cpath = pjoin(self.location, category.lstrip(os.path.sep))
        l = set()
        d = {}
        bad = False
        try:
            for x in listdir_dirs(cpath):
                if x.startswith(".tmp.") or x.endswith(".lockfile") \
                        or x.startswith("-MERGING-"):
                    continue
                try:
                    pkg = VersionedCPV(f'{category}/{x}')
                except InvalidCPV:
                    bad = True
                if bad or not pkg.fullver:
                    if '-scm' in x:
                        bad = 'scm'
                    elif '-try' in x:
                        bad = 'try'
                    else:
                        raise InvalidCPV(f'{category}/{x}', 'no version component')
                    logger.error(
                        f'merged -{bad} pkg detected: {category}/{x}. '
                        f'throwing exception due to -{bad} not being a valid'
                        ' version component.  Silently ignoring that '
                        'specific version is not viable either since it '
                        'would result in pkgcore stomping whatever it was '
                        f'that -{bad} version merged.  '
                        'Use the offending pkg manager that merged it to '
                        'unmerge it.')
                    raise InvalidCPV(
                        f'{category}/{x}', f'{bad} version component is not standard.')
                l.add(pkg.package)
                d.setdefault((category, pkg.package), []).append(pkg.fullver)
        except EnvironmentError as e:
            category = pjoin(self.location, category.lstrip(os.path.sep))
            raise KeyError(f'failed fetching packages for category {category}: {e}') from e

        self._versions_tmp_cache.update(d)
        return tuple(l)

    def _get_versions(self, catpkg):
        return tuple(self._versions_tmp_cache.pop(catpkg))

    def _get_ebuild_path(self, pkg):
        s = f"{pkg.package}-{pkg.fullver}"
        return pjoin(self.location, pkg.category, s, s + ".ebuild")

    def _get_path(self, pkg):
        s = f"{pkg.package}-{pkg.fullver}"
        return pjoin(self.location, pkg.category, s)

    _metadata_rewrites = {
        "bdepend": "BDEPEND", "depend": "DEPEND", "rdepend": "RDEPEND", "pdepend": "PDEPEND",
        "idepend": "IDEPEND",
        "use": "USE", "eapi": "EAPI", "CONTENTS": "contents",
        "source_repository": "repository", "fullslot": "SLOT",
    }

    def _get_metadata(self, pkg):
        return IndeterminantDict(
            partial(self._internal_load_key, pjoin(
                self.location, pkg.category,
                f"{pkg.package}-{pkg.fullver}")))

    def _internal_load_key(self, path, key):
        key = self._metadata_rewrites.get(key, key)
        if key == "contents":
            data = ContentsFile(pjoin(path, "CONTENTS"), mutable=True)
        elif key == "environment":
            fp = pjoin(path, key)
            if not os.path.exists(f'{fp}.bz2'):
                if not os.path.exists(fp):
                    # icky.
                    raise KeyError("environment: no environment file found")
                data = data_source.local_source(fp)
            else:
                data = data_source.bz2_source(f'{fp}.bz2')
        elif key == 'ebuild':
            fp = pjoin(path, os.path.basename(path.rstrip(os.path.sep)) + '.ebuild')
            data = data_source.local_source(fp)
        elif key == 'repo':
            # try both, for portage/paludis compatibility.
            data = readfile(pjoin(path, 'repository'), True)
            if data is None:
                data = readfile(pjoin(path, 'REPOSITORY'), True)
                if data is None:
                    raise KeyError(key)
        else:
            data = readfile(pjoin(path, key), True)
            if data is None:
                raise KeyError((path, key))
            data = data.rstrip('\n')
        return data

    def notify_remove_package(self, pkg):
        remove_it = len(self.packages[pkg.category]) == 1
        prototype.tree.notify_remove_package(self, pkg)
        if remove_it:
            try:
                os.rmdir(pjoin(self.location, pkg.category))
            except OSError as oe:
                # POSIX specifies either ENOTEMPTY or EEXIST for non-empty dir
                # in particular, Solaris uses EEXIST in that case.
                # https://github.com/pkgcore/pkgcore/pull/181
                if oe.errno not in (errno.ENOTEMPTY, errno.EEXIST):
                    raise
                # silently swallow it;
                del oe

    def __str__(self):
        return f"{self.repo_id}: location {self.location}"


class _WrappedInstalledPkg(pkg_base.wrapper):
    """Installed package with configuration data bound to it."""

    built = True
    __slots__ = ()

    def __str__(self):
        return (
            f'installed pkg: {self.cpvstr}::{self.repo.repo_id}, '
            f'source repo {self.source_repository!r}'
        )


class ConfiguredTree(wrapper.tree, tree):
    """Configured repository for packages installed on the filesystem."""

    configured = True
    frozen_settable = False

    def __init__(self, vdb, domain, domain_settings):
        _WrappedInstalledPkg._operations = self._generate_operations
        _WrappedInstalledPkg.repo = self
        wrapper.tree.__init__(self, vdb, package_class=_WrappedInstalledPkg)
        self.domain = domain
        self.domain_settings = domain_settings

    def _generate_operations(self, domain, pkg, **kwargs):
        pkg = pkg._raw_pkg
        return ebd.built_operations(
            domain, pkg, initial_env=self.domain_settings, **kwargs)



__all__ = ("Installed", "VersionedInstalled")

import operator

from ..config.hint import ConfigHint
from ..restrictions import packages, values


class _Base:
    """Base for Installed and VersionedInstalled."""

    def __init__(self, vdb):
        self.vdbs = vdb

    def __iter__(self):
        restrict = packages.PackageRestriction(
            "package_is_real", values.EqualityMatch(True))
        for repo in self.vdbs:
            for pkg in repo.itermatch(restrict):
                yield self.getter(pkg)


class Installed(_Base):
    """Set of packages holding slotted atoms of all installed packages."""
    pkgcore_config_type = ConfigHint({'vdb': 'refs:repo'}, typename='pkgset')
    getter = operator.attrgetter('slotted_atom')


class VersionedInstalled(_Base):
    """Set of packages holding versioned atoms of all installed packages."""
    pkgcore_config_type = ConfigHint({'vdb': 'refs:repo'}, typename='pkgset')
    getter = operator.attrgetter('versioned_atom')



__all__ = ("sqfs_syncer",)

from . import base
from .http import http_syncer


class sqfs_syncer(http_syncer):

    supported_uris = (
        ('sqfs+http://', 5),
        ('sqfs+https://', 5),
    )

    @staticmethod
    def parse_uri(raw_uri):
        if raw_uri.startswith(("sqfs+http://", "sqfs+https://")):
            return raw_uri[5:]
        raise base.UriError(raw_uri, "unsupported URI")

    def _post_download(self, path):
        # TODO: verify gpg data if it exists
        super()._post_download(path)



from pkgcore.config import basics
from pkgcore.config.hint import ConfigHint
from pkgcore.scripts import pclonecache
from pkgcore.test.scripts.helpers import ArgParseMixin
from snakeoil.test import TestCase


class Cache:

    pkgcore_config_type = ConfigHint(typename='cache')

    def __init__(self, readonly=True):
        self.readonly = self.frozen = readonly


class CommandlineTest(TestCase, ArgParseMixin):

    _argparser = pclonecache.argparser

    def test_parser(self):
        self.assertError(
            'the following arguments are required: target',
            'spork')
        self.assertError(
            "argument source: couldn't find cache 'spork'",
            'spork', 'spork2')
        self.assertError(
            "argument target: couldn't find cache 'spork2' (available: spork)",
            'spork', 'spork2',
            spork=basics.HardCodedConfigSection({'class': Cache}))
        self.assertError(
            "argument target: cache 'spork2' is readonly",
            'spork', 'spork2',
            spork=basics.HardCodedConfigSection({'class': Cache,}),
            spork2=basics.HardCodedConfigSection({'class': Cache,}))
